"""
DeviceSetup models for lara_django_material.

This module contains DeviceSetup-related models including shape, DeviceSetup,
and subset classes.
"""

import uuid

from django.conf import settings
from django.db import models
from django.utils import timezone
from lara_django_base.models import MediaType
from lara_django_people.models import ItemSubsetAbstr

from .base import ExtraData, PartAbstr, PartDeviceClass, PartShapeAbstr
from .devices import Device


class DeviceSetupShape(PartShapeAbstr):
    """DeviceSetup shape and dimensions."""

    model_curi = "lara:material/DeviceSetupShape"
    devicesetupshape_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    data_extra = models.ManyToManyField(
        ExtraData,
        related_name="%(app_label)s_%(class)s_dss_data_extra_related",
        related_query_name="%(app_label)s_%(class)s_dss_data_extra_related_query",
        blank=True,
        help_text="part shape extra data, like 3D model or polygon describing the part shape/dimensions",
    )

    def save(self, *args, **kwargs):
        """Generate default values for IRI."""
        if self.iri is None or self.iri == "":
            self.iri = f"{settings.LARA_PREFIXES['lara']}material/DeviceSetupShape/{self.name.replace(' ', '_').lower().strip()}"

        super().save(*args, **kwargs)


class DeviceSetup(PartAbstr):
    """
    Setup or assembly of devices.

    E.g., setup of a physical or chemical experiment: distillation apparatus,
    flow reactor, ultra-high vacuum depositor, ... with all connectivities.
    """

    model_curi = "lara:material/DeviceSetup"
    devicesetup_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    setup_class = models.ForeignKey(
        PartDeviceClass,
        related_name="%(app_label)s_%(class)s_setup_classes_related",
        related_query_name="%(app_label)s_%(class)s_setup_classes_related_query",
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        help_text="screw, bolt, nut, lamp, diode, resistor, IC, ...",
    )
    shape = models.ForeignKey(
        DeviceSetupShape,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        help_text="shape of Device setup",
    )
    components = models.ManyToManyField(
        Device,
        related_query_name="%(app_label)s_%(class)s_components_related_query",
        blank=True,
        help_text="components of the device setup",
    )
    blueprint_image = models.ImageField(
        upload_to="material/",
        blank=True,
        null=True,
        # default="",
        help_text="rel. path/filename to image",
    )
    blueprint_media_type = models.ForeignKey(
        MediaType,
        related_name="%(app_label)s_%(class)s_media_types_related",
        related_query_name="%(app_label)s_%(class)s_media_types_related_query",
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        help_text="file type of extra data file",
    )
    blueprint_file = models.FileField(upload_to="material/", blank=True, null=True, help_text="blueprint file")
    blueprint_pdf = models.FileField(upload_to="material/", blank=True, null=True, help_text="blueprint PDF file")
    scheme_svg = models.TextField(blank=True, null=True, help_text="setup scheme in SVG")
    assembly = models.JSONField(blank=True, null=True, help_text="assembly with connections in JSON")
    data_extra = models.ManyToManyField(
        ExtraData,
        related_name="%(app_label)s_%(class)s_ds_data_extra_related",
        related_query_name="%(app_label)s_%(class)s_ds_data_extra_related_query",
        blank=True,
        help_text="e.g. manual",
    )

    def save(self, *args, **kwargs):
        """Generate default values for type_barcode and IRI."""
        if self.name is None or self.name == "":
            if self.name_display is not None or self.name_display != "":
                self.name = self.name_display.replace(" ", "_").lower().strip()

        if self.name is not None or self.name != "":
            self.name = self.name.replace(" ", "_").lower().strip()

        if self.name_display is None or self.name_display == "":
            self.name_display = self.name

        if self.type_barcode == "" or self.type_barcode is None:
            self.type_barcode = f"AUTO_{str(uuid.uuid4())[:8]}"  # avoiding unique constraint violation

        if self.iri is None or self.iri == "":
            self.iri = f"{settings.LARA_PREFIXES['lara']}material/Device/{self.name.replace(' ', '_').lower().strip()}_{str(self.part_id)[:8]}"

        if self.datetime_last_modified is None:
            self.datetime_last_modified = timezone.now()

        super().save(*args, **kwargs)


class DeviceSetupsSubset(ItemSubsetAbstr):
    """
    DeviceSetups subset.

    It can be used for user or group defined subsets of device setups.
    """

    model_curi = "lara:material/DeviceSetupsSubset"
    devicesetupssubset_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    devicesetups = models.ManyToManyField(
        DeviceSetup,
        related_name="%(app_label)s_%(class)s_devicesetups_subset_related",
        related_query_name="%(app_label)s_%(class)s_devicesetups_subset_related_query",
        blank=True,
        help_text="device setups in the subset",
        through="OrderedDeviceSetupsSubset",
    )


class OrderedDeviceSetupsSubset(models.Model):
    """
    Through model for DeviceSetupsSubset and DeviceSetup relationship.

    This class can be used to store the order of device setups in a subset.
    """

    model_curi = "lara:material/OrderedDeviceSetupsSubset"

    ordereddevicessubset_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    devicesetup = models.ForeignKey(
        DeviceSetup,
        related_name="%(app_label)s_%(class)s_devicesetup_related",
        related_query_name="%(app_label)s_%(class)s_devicesetup",
        on_delete=models.CASCADE,
        help_text="device setup in the subset",
    )

    devicesetups_subset = models.ForeignKey(
        DeviceSetupsSubset,
        related_name="%(app_label)s_%(class)s_devicesetups_subset_related",
        related_query_name="%(app_label)s_%(class)s_devicesetups_subset",
        on_delete=models.CASCADE,
        help_text="subset of device setups",
    )

    order = models.IntegerField(help_text="order of device setup in subset")

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["devicesetup", "devicesetups_subset"],
                name="%(app_label)s_%(class)s_devicesetup_devicesetups_subset_unique",
            ),
        ]
        ordering = ("order",)



"""
Base forms module for lara_django_material.

:PROJECT: LARAsuite

:details: Common form utilities, widgets, and renderer
         optimized for Python 3.13+ with ClassVar pattern.

:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: Contains shared components used across all form modules
.. todo:: -
"""

from typing import ClassVar

from django import forms
from django.forms.renderers import TemplatesSetting

from ..models import ExtraData, PartDeviceClass


class CustomFormRenderer(TemplatesSetting):
    """Custom form renderer for lara_django_material forms."""

    field_template_name = "lara_django_material/forms/field_group.html"


class SearchableSelectMultiple(forms.SelectMultiple):
    """Custom widget for searchable multi-select using Alpine.js."""

    template_name = "lara_django_material/widgets/searchable_select_multiple.html"

    def __init__(self, attrs=None):
        """Initialize widget with custom CSS class."""
        default_attrs = {"class": "searchable-select-multiple"}
        if attrs:
            default_attrs.update(attrs)
        super().__init__(attrs=default_attrs)


## -------------- ExtraData -------------- ##


class ExtraDataCreateForm(forms.ModelForm):
    """Form for creating new ExtraData entries."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:
        """Meta configuration for ExtraDataCreateForm."""

        model = ExtraData
        fields = (
            "data_type",
            "namespace",
            "name",
            "name_display",
            "version",
            "url",
            "description",
            "file",
            "media_type",
        )

        text_fields_required: ClassVar = ("data_type", "name")
        text_fields: ClassVar = ("name_display", "version", "media_type")
        textarea_fields: ClassVar = ("description",)
        url_fields: ClassVar = ("url",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "namespace": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "file": forms.ClearableFileInput(attrs={"class": "file-input file-input-sm file-input-bordered w-full"}),
        }


class ExtraDataUpdateForm(forms.ModelForm):
    """Form for updating existing ExtraData entries."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:
        """Meta configuration for ExtraDataUpdateForm."""

        model = ExtraData
        fields = (
            "data_type",
            "name",
            "name_display",
            "namespace",
            "version",
            "hash_sha256",
            "data_json",
            "media_type",
            "iri",
            "url",
            "description",
            "image",
            "file",
        )

        text_fields_required: ClassVar = ("data_type", "name")
        text_fields: ClassVar = ("name_display", "version", "media_type", "hash_sha256")
        textarea_fields: ClassVar = ("description",)
        url_fields: ClassVar = ("iri", "url")

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in text_fields
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
            "namespace": forms.Select(attrs={"class": "select select-sm select-bordered w-full"}),
            "data_json": forms.Textarea(
                attrs={"class": "textarea textarea-sm textarea-bordered w-full font-mono", "rows": 6}
            ),
            "image": forms.ClearableFileInput(attrs={"class": "file-input file-input-sm file-input-bordered w-full"}),
            "file": forms.ClearableFileInput(attrs={"class": "file-input file-input-sm file-input-bordered w-full"}),
        }


## -------------- PartDeviceClass -------------- ##


class PartDeviceClassCreateForm(forms.ModelForm):
    """Form for creating new PartDeviceClass entries."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:
        """Meta configuration for PartDeviceClassCreateForm."""

        model = PartDeviceClass
        fields = ("name", "iri", "description")

        text_fields_required: ClassVar = ("name",)
        textarea_fields: ClassVar = ("description",)
        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
        }


class PartDeviceClassUpdateForm(forms.ModelForm):
    """Form for updating existing PartDeviceClass entries."""

    default_renderer: ClassVar = CustomFormRenderer()

    class Meta:
        """Meta configuration for PartDeviceClassUpdateForm."""

        model = PartDeviceClass
        fields = ("name", "iri", "description")

        text_fields_required: ClassVar = ("name",)
        textarea_fields: ClassVar = ("description",)
        url_fields: ClassVar = ("iri",)

        widgets: ClassVar = {
            **{
                field: forms.TextInput(attrs={"class": "input input-sm input-bordered w-full", "required": True})
                for field in text_fields_required
            },
            **{
                field: forms.Textarea(attrs={"class": "textarea textarea-sm textarea-bordered w-full", "rows": 3})
                for field in textarea_fields
            },
            **{field: forms.URLInput(attrs={"class": "input input-sm input-bordered w-full"}) for field in url_fields},
        }



"""
_____________________________________________________________________.

:PROJECT: LARAsuite

*lara_django_material admin *

:details: lara_django_material admin module admin backend configuration.
         -
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: - run "lara-django-dev admin_generator lara_django_material >> admin.py" to update this file
________________________________________________________________________
"""

# -*- coding: utf-8 -*-
from django.contrib import admin

from .models import (
    Device,
    DeviceSetup,
    DeviceSetupShape,
    DeviceShape,
    ExtraData,
    Labware,
    LabwareClass,
    LabwareShape,
    Part,
    PartDeviceClass,
    PartShape,
    WellShape,
)


@admin.register(ExtraData)
class ExtraDataAdmin(admin.ModelAdmin):
    """ExtraDataAdmin class."""

    list_display = (
        "name",
        "data_type",
        "namespace",
        "url",
        "description",
        "file",
        "media_type",
        "extradata_id",
    )
    list_filter = ("data_type", "namespace", "media_type")


@admin.register(PartDeviceClass)
class PartDeviceClassAdmin(admin.ModelAdmin):
    """PartDeviceClassAdmin class."""

    list_display = (
        "name",
        "description",
        "partdeviceclass_id",
    )
    search_fields = ("name",)


@admin.register(PartShape)
class PartShapeAdmin(admin.ModelAdmin):
    """PartShapeAdmin class."""

    list_display = (
        "name",
        "name_standard",
        "width",
        "length",
        "height_unlidded",
        "hight_lidded",
        "hight_stacked",
        "hight_stacked_lidded",
        "description",
        "partshape_id",
    )
    raw_id_fields = ("data_extra",)
    search_fields = ("name",)


@admin.register(Part)
class PartAdmin(admin.ModelAdmin):
    """PartAdmin class."""

    list_display = (
        "name",
        "product_type",
        "type_barcode",
        "model_no",
        "product_no",
        "energy_consumption",
        "part_class",
        "part_id",
    )
    list_filter = ("manufacturer", "part_class", "shape")
    raw_id_fields = ("tags", "data_extra")
    search_fields = ("name",)


@admin.register(DeviceShape)
class DeviceShapeAdmin(admin.ModelAdmin):
    """DeviceShapeAdmin class."""

    list_display = (
        "name",
        "name_standard",
        "width",
        "length",
        "height_unlidded",
        "hight_lidded",
        "hight_stacked",
        "description",
        "deviceshape_id",
    )
    raw_id_fields = ("data_extra",)
    search_fields = ("name",)


@admin.register(Device)
class DeviceAdmin(admin.ModelAdmin):
    """DeviceAdmin class."""

    list_display = (
        "name",
        "product_type",
        "type_barcode",
        "model_no",
        "product_no",
        "energy_consumption",
        "device_class",
        "device_id",
    )
    list_filter = (
        "manufacturer",
        "device_class",
    )
    raw_id_fields = ("tags", "data_extra", "components")
    search_fields = ("name",)


@admin.register(DeviceSetupShape)
class DeviceSetupShapeAdmin(admin.ModelAdmin):
    """DeviceSetupShapeAdmin class."""

    list_display = (
        "name",
        "name_standard",
        "width",
        "length",
        "height_unlidded",
        "hight_lidded",
        "hight_stacked",
        "hight_stacked_lidded",
        "description",
        "devicesetupshape_id",
    )
    raw_id_fields = ("data_extra",)
    search_fields = ("name",)


@admin.register(DeviceSetup)
class DeviceSetupAdmin(admin.ModelAdmin):
    """DeviceSetupAdmin class."""

    list_display = (
        "name",
        "setup_class",
        "url",
        "model_no",
        "product_type",
        "type_barcode",
        "product_no",
        "weight",
        "energy_consumption",
        "description",
        "devicesetup_id",
    )
    list_filter = (
        "manufacturer",
        "setup_class",
        "shape",
        "blueprint_media_type",
    )
    raw_id_fields = ("tags", "components", "data_extra")
    search_fields = ("name",)


@admin.register(LabwareClass)
class LabwareClassAdmin(admin.ModelAdmin):
    """LabwareClassAdmin class."""

    list_display = ("labwareclass_id", "name", "iri", "description")
    search_fields = ("name",)


@admin.register(WellShape)
class WellShapeAdmin(admin.ModelAdmin):
    """WellShapeAdmin class."""

    list_display = (
        "name",
        "uri",
        "iri",
        "depth_well",
        "shape_well",
        "shape_bottom",
        "top_radius_xy",
        "bottom_radius_xy",
        "bottom_radius_z",
        "cone_angle",
        "cone_depth",
        "wellshape_id",
    )
    raw_id_fields = ("data_extra",)
    search_fields = ("name",)


@admin.register(LabwareShape)
class LabwareShapeAdmin(admin.ModelAdmin):
    """LabwareShapeAdmin class."""

    list_display = (
        "name",
        "name_standard",
        "width",
        "length",
        "height_unlidded",
        "hight_lidded",
        "hight_stacked",
        "hight_stacked_lidded",
        "well_dist_row",
        "well_dist_col",
        "well_shape",
        "description",
        "labwareshape_id",
    )
    list_filter = ("well_shape",)
    raw_id_fields = ("data_extra",)
    search_fields = ("name",)


@admin.register(Labware)
class LabwareAdmin(admin.ModelAdmin):
    """LabwareAdmin class."""

    list_display = (
        "name",
        "labware_class",
        "product_type",
        "type_barcode",
        "model_no",
        "product_no",
        "num_rows",
        "num_cols",
        "num_wells",
        "shape",
        "liddable",
        "sealable",
        "labware_id",
    )
    list_filter = (
        "manufacturer",
        "labware_class",
        "shape",
        "liddable",
        "sealable",
    )
    raw_id_fields = ("tags", "lids", "data_extra")
    search_fields = ("name",)



"""
_____________________________________________________________________.

:PROJECT: LARAsuite

*lara_django_myproj gRPC services*

:details: lara_django_myproj gRPC services.
         -
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: -
________________________________________________________________________
"""

## generated with django-socio-grpc generateprpcinterface lara_django_material  (LARA-version)

import lara_django_material_grpc.v1.lara_django_material_pb2 as lara_django_material_pb2
from django_socio_grpc import generics, mixins
from lara_django_base.grpc.mixins import AsyncDownloadModelMixin, AsyncRetrieveByNameModelMixin, AsyncUploadModelMixin

from lara_django_material.models import (
    Device,
    DeviceSetup,
    DeviceSetupShape,
    DeviceSetupsSubset,
    DeviceShape,
    DevicesSubset,
    ExtraData,
    Labware,
    LabwareClass,
    LabwareShape,
    LabwareSubset,
    Part,
    PartDeviceClass,
    PartShape,
    PartsSubset,
    WellShape,
)

from ..filters import (
    DeviceFilterSet,
    DeviceSetupFilterSet,
    DeviceSetupShapeFilterSet,
    DeviceShapeFilterSet,
    ExtraDataFilterSet,
    LabwareClassFilterSet,
    LabwareFilterSet,
    LabwareShapeFilterSet,
    PartDeviceClassFilterSet,
    PartFilterSet,
    PartShapeFilterSet,
    WellShapeFilterSet,
)
from .serializers import (
    DeviceProtoSerializer,
    DeviceSetupProtoSerializer,
    DeviceSetupShapeProtoSerializer,
    DeviceSetupsSubsetProtoSerializer,
    DeviceShapeProtoSerializer,
    DevicesSubsetProtoSerializer,
    ExtraDataProtoSerializer,
    LabwareClassProtoSerializer,
    LabwareProtoSerializer,
    LabwareShapeProtoSerializer,
    LabwareSubsetProtoSerializer,
    PartDeviceClassByNameProtoSerializer,
    PartDeviceClassProtoSerializer,
    PartProtoSerializer,
    PartShapeProtoSerializer,
    PartsSubsetProtoSerializer,
    WellShapeProtoSerializer,
)


class ExtraDataService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    """ExtraDataService class."""

    queryset = ExtraData.objects.all()
    serializer_class = ExtraDataProtoSerializer
    filterset_class = ExtraDataFilterSet
    search_fields = ["name", "description"]


class PartDeviceClassService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    """PartDeviceClassService class."""

    queryset = PartDeviceClass.objects.all()
    serializer_class = PartDeviceClassProtoSerializer
    filterset_class = PartDeviceClassFilterSet
    search_fields = ["name", "description"]


class PartDeviceClassByNameService(
    generics.GenericService,
    mixins.AsyncListModelMixin,
    AsyncUploadModelMixin,
    AsyncRetrieveByNameModelMixin,
    AsyncDownloadModelMixin,
):
    """PartDeviceClassByNameService class."""

    pb2 = lara_django_material_pb2
    queryset = PartDeviceClass.objects.all().order_by("name")
    serializer_class = PartDeviceClassByNameProtoSerializer
    filterset_class = PartDeviceClassFilterSet

    lookup_field = "name"


class PartShapeService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    """PartShapeService class."""

    queryset = PartShape.objects.all()
    serializer_class = PartShapeProtoSerializer
    filterset_class = PartShapeFilterSet
    search_fields = ["name", "description"]


class PartService(
    generics.AsyncModelService,
    mixins.AsyncStreamModelMixin,
    AsyncUploadModelMixin,
    AsyncRetrieveByNameModelMixin,
    AsyncDownloadModelMixin,
):
    """PartService class."""

    pb2 = lara_django_material_pb2
    queryset = Part.objects.all()
    serializer_class = PartProtoSerializer
    filterset_class = PartFilterSet
    search_fields = ["name", "description"]


# class PartByNameService(generics.GenericService, mixins.AsyncListModelMixin, mixins.AsyncRetrieveModelMixin):
#     queryset = Part.objects.all().order_by("name")
#     serializer_class = PartByNameProtoSerializer
#     filterset_class = PartFilterSet

#     lookup_field = "name"


class PartsSubsetService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    """PartsSubsetService class."""

    queryset = PartsSubset.objects.all()
    serializer_class = PartsSubsetProtoSerializer
    search_fields = ["name", "description"]


class DeviceShapeService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    """DeviceShapeService class."""

    queryset = DeviceShape.objects.all()
    serializer_class = DeviceShapeProtoSerializer
    filterset_class = DeviceShapeFilterSet
    search_fields = ["name", "description"]


class DeviceService(
    generics.AsyncModelService,
    mixins.AsyncStreamModelMixin,
    AsyncUploadModelMixin,
    AsyncRetrieveByNameModelMixin,
    AsyncDownloadModelMixin,
):
    """DeviceService class."""

    pb2 = lara_django_material_pb2
    queryset = Device.objects.all()
    serializer_class = DeviceProtoSerializer
    filterset_class = DeviceFilterSet
    search_fields = ["name", "description"]


# class DeviceByNameService(generics.GenericService, mixins.AsyncListModelMixin, mixins.AsyncRetrieveModelMixin):
#     queryset = Device.objects.all().order_by("name")
#     serializer_class = DeviceByNameProtoSerializer
#     filterset_class = DeviceFilterSet

#     lookup_field = "name"


class DevicesSubsetService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    """DevicesSubsetService class."""

    queryset = DevicesSubset.objects.all()
    serializer_class = DevicesSubsetProtoSerializer
    search_fields = ["name", "description"]


class DeviceSetupShapeService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    """DeviceSetupShapeService class."""

    queryset = DeviceSetupShape.objects.all()
    serializer_class = DeviceSetupShapeProtoSerializer
    filterset_class = DeviceSetupShapeFilterSet
    search_fields = ["name", "description"]


class DeviceSetupService(
    generics.AsyncModelService,
    mixins.AsyncStreamModelMixin,
    AsyncUploadModelMixin,
    AsyncRetrieveByNameModelMixin,
    AsyncDownloadModelMixin,
):
    """DeviceSetupService class."""

    pb2 = lara_django_material_pb2
    queryset = DeviceSetup.objects.all()
    serializer_class = DeviceSetupProtoSerializer
    filterset_class = DeviceSetupFilterSet
    search_fields = ["name", "description"]


# class DeviceSetupByNameService(generics.GenericService, mixins.AsyncListModelMixin, mixins.AsyncRetrieveModelMixin):
#     queryset = DeviceSetup.objects.all().order_by("name")
#     serializer_class = DeviceSetupByNameProtoSerializer
#     filterset_class = DeviceSetupFilterSet

#     lookup_field = "name"


class DeviceSetupsSubsetService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    """DeviceSetupsSubsetService class."""

    queryset = DeviceSetupsSubset.objects.all()
    serializer_class = DeviceSetupsSubsetProtoSerializer
    search_fields = ["name", "description"]


class LabwareClassService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    """LabwareClassService class."""

    queryset = LabwareClass.objects.all()
    serializer_class = LabwareClassProtoSerializer
    filterset_class = LabwareClassFilterSet
    search_fields = ["name", "description"]


class WellShapeService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    """WellShapeService class."""

    queryset = WellShape.objects.all()
    serializer_class = WellShapeProtoSerializer
    filterset_class = WellShapeFilterSet
    search_fields = ["name", "description"]


class LabwareShapeService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    """LabwareShapeService class."""

    queryset = LabwareShape.objects.all()
    serializer_class = LabwareShapeProtoSerializer
    filterset_class = LabwareShapeFilterSet
    search_fields = ["name", "description"]


class LabwareService(
    generics.AsyncModelService,
    mixins.AsyncStreamModelMixin,
    AsyncUploadModelMixin,
    AsyncRetrieveByNameModelMixin,
    AsyncDownloadModelMixin,
):
    """LabwareService class."""

    pb2 = lara_django_material_pb2
    queryset = Labware.objects.all()
    serializer_class = LabwareProtoSerializer
    filterset_class = LabwareFilterSet
    search_fields = ["name", "description"]


# class LabwareByNameService(generics.GenericService, mixins.AsyncListModelMixin, mixins.AsyncRetrieveModelMixin):
#     queryset = Labware.objects.all().order_by("name")
#     serializer_class = LabwareByNameProtoSerializer
#     filterset_class = LabwareFilterSet

#     lookup_field = "name"


class LabwareSubsetService(generics.AsyncModelService, mixins.AsyncStreamModelMixin):
    """LabwareSubsetService class."""

    queryset = LabwareSubset.objects.all()
    serializer_class = LabwareSubsetProtoSerializer
    search_fields = ["name", "description"]



"""
Devices models for lara_django_material.

This module contains Device-related models including shape, Device,
and subset classes.
"""

import uuid

from django.conf import settings
from django.db import models
from django.utils import timezone
from lara_django_people.models import ItemSubsetAbstr

from .base import ExtraData, PartAbstr, PartDeviceClass, PartShapeAbstr
from .parts import Part


class DeviceShape(PartShapeAbstr):
    """Device shape and dimensions."""

    model_curi = "lara:material/DeviceShape"
    deviceshape_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    data_extra = models.ManyToManyField(
        ExtraData,
        related_name="%(app_label)s_%(class)s_dev_shape_data_extra_related",
        related_query_name="%(app_label)s_%(class)s_dev_shape_data_extra_related_query",
        blank=True,
        help_text="device shape extra data, like 3D model or polygon describing the device shape/dimensions",
    )

    def save(self, *args, **kwargs):
        """Generate default values for IRI."""
        if self.iri is None or self.iri == "":
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}material/DeviceShape/{self.name.replace(' ', '_').lower().strip()}"
            )

        super().save(*args, **kwargs)


class Device(PartAbstr):
    """Generic device class."""

    model_curi = "lara:material/Device"
    device_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    device_class = models.ForeignKey(
        PartDeviceClass,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        help_text="UV-spectrometer, MALDI-TOF-MS, HPLC, ..",
    )
    labware_min_capacity = models.IntegerField(
        null=True,
        blank=True,
        help_text="labware capacity, min. number of labware, some centrifuges need at least 2 labware, e.g. for balancing",
    )
    labware_max_capacity = models.IntegerField(
        null=True, blank=True, help_text="labware capacity, max. number of labware"
    )
    labware_max_height = models.FloatField(null=True, blank=True, help_text="max. labware height")
    labware_storage_layout = models.JSONField(
        null=True,
        blank=True,
        help_text="labware storage layout, like num. of stackers, num. of labware per stacker, orientation of nests, heights, ... ",
    )
    shape = models.ForeignKey(
        DeviceShape,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        help_text="shape of Device",
    )
    components = models.ManyToManyField(
        Part,
        related_query_name="%(app_label)s_%(class)s_components_related_query",
        blank=True,
        help_text="components of the device setup",
    )
    data_extra = models.ManyToManyField(
        ExtraData,
        related_name="%(app_label)s_%(class)s_devices_data_extra_related",
        related_query_name="%(app_label)s_%(class)s_devices_data_extra_related_query",
        blank=True,
        help_text="e.g. manual",
    )

    def save(self, *args, **kwargs):
        """Generate default values for type barcode and IRI."""
        if self.name is None or self.name == "":
            if self.name_display is not None or self.name_display != "":
                self.name = self.name_display.replace(" ", "_").lower().strip()

        if self.name is not None or self.name != "":
            self.name = self.name.replace(" ", "_").lower().strip()

        if self.name_display is None or self.name_display == "":
            self.name_display = self.name

        if self.type_barcode == "" or self.type_barcode is None:
            self.type_barcode = f"AUTO_{str(uuid.uuid4())[:8]}"  # avoiding unique constraint violation

        if self.iri is None or self.iri == "":
            self.iri = f"{settings.LARA_PREFIXES['lara']}material/Device/{self.name.replace(' ', '_').lower().strip()}_{str(self.device_id)[:8]}"

        if self.datetime_last_modified is None:
            self.datetime_last_modified = timezone.now()

        super().save(*args, **kwargs)


class DevicesSubset(ItemSubsetAbstr):
    """
    Devices subset.

    It can be used for user or group defined subsets of devices.
    """

    model_curi = "lara:material/DevicesSubset"
    devicessubset_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    devices = models.ManyToManyField(
        Device,
        related_name="%(app_label)s_%(class)s_devices_subset_related",
        related_query_name="%(app_label)s_%(class)s_devices_subset_related_query",
        blank=True,
        help_text="devices in the subset",
        through="OrderedDevicesSubset",
    )


class OrderedDevicesSubset(models.Model):
    """
    Through model for DevicesSubset and Device relationship.

    This class can be used to store the order of devices in a subset.
    """

    model_curi = "lara:material/OrderedDevicesSubset"

    ordereddevicessubset_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    device = models.ForeignKey(
        Device,
        related_name="%(app_label)s_%(class)s_device_related",
        related_query_name="%(app_label)s_%(class)s_device",
        on_delete=models.CASCADE,
        help_text="device in the subset",
    )

    devices_subset = models.ForeignKey(
        DevicesSubset,
        related_name="%(app_label)s_%(class)s_devices_subset_related",
        related_query_name="%(app_label)s_%(class)s_devices_subset",
        on_delete=models.CASCADE,
        help_text="subset of devices",
    )

    order = models.IntegerField(help_text="order of device in subset")

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["device", "devices_subset"],
                name="%(app_label)s_%(class)s_device_devices_subset_unique",
            ),
        ]
        ordering = ("order",)



"""
_____________________________________________________________________.

:PROJECT: LARAsuite

*lara_django_data filter app *

:details: lara_django_data filter app.
         -
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: -
________________________________________________________________________
"""

from django_filters.rest_framework import FilterSet

from .models import (
    Device,
    DeviceSetup,
    DeviceSetupShape,
    DeviceShape,
    ExtraData,
    Labware,
    LabwareClass,
    LabwareShape,
    Part,
    PartDeviceClass,
    PartShape,
    WellShape,
)


class ExtraDataFilterSet(FilterSet):
    """ExtraDataFilterSet class."""

    class Meta:
        """Meta class."""

        model = ExtraData
        fields = {
            "name_display": ["exact", "contains"],
            "name": ["exact", "contains"],
            "iri": ["exact", "contains"],
            "url": ["exact", "contains"],
            "datetime_created": ["lt", "gt", "exact", "range"],
            "datetime_last_modified": ["lt", "gt", "exact", "range"],
            "description": ["exact", "contains"],
        }


class PartDeviceClassFilterSet(FilterSet):
    """PartDeviceClassFilterSet class."""

    class Meta:
        """Meta class."""

        model = PartDeviceClass
        fields = {
            "name": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class PartShapeFilterSet(FilterSet):
    """PartShapeFilterSet class."""

    class Meta:
        """Meta class."""

        model = PartShape
        fields = {
            "name": ["exact", "contains"],
            "description": ["exact", "contains"],
        }


class PartFilterSet(FilterSet):
    """PartFilterSet class."""

    model = Part
    fields = {
        "name": ["exact", "contains"],
        "name_display": ["exact", "contains"],
        "url": ["exact", "contains"],
        "pid": ["exact", "contains"],
        "iri": ["exact", "contains"],
        "model_no": ["exact", "contains"],
        "product_type": ["exact", "contains"],
        "type_barcode": ["exact", "contains"],
        "product_no": ["exact", "contains"],
        "weight": ["lt", "gt", "exact", "range"],
        "energy_consumption": ["lt", "gt", "exact", "range"],
        "datetime_last_modified": ["lt", "gt", "exact", "range"],
        "description": ["exact", "contains"],
    }


class DeviceShapeFilterSet(FilterSet):
    """DeviceShapeFilterSet class."""

    model = DeviceShape
    fields = {
        "name": ["exact", "contains"],
        "description": ["exact", "contains"],
    }


class DeviceFilterSet(FilterSet):
    """DeviceFilterSet class."""

    model = Device
    fields = {
        "name": ["exact", "contains"],
        "name_display": ["exact", "contains"],
        "manufacturer__name": ["exact", "contains"],
        "url": ["exact", "contains"],
        "pid": ["exact", "contains"],
        "iri": ["exact", "contains"],
        "model_no": ["exact", "contains"],
        "product_type": ["exact", "contains"],
        "type_barcode": ["exact", "contains"],
        "product_no": ["exact", "contains"],
        "weight": ["lt", "gt", "exact", "range"],
        "energy_consumption": ["lt", "gt", "exact", "range"],
        "datetime_last_modified": ["lt", "gt", "exact", "range"],
        "description": ["exact", "contains"],
        "labware_min_capacity": ["lt", "gt", "exact"],
        "labware_max_capacity": ["lt", "gt", "exact"],
        "labware_max_height": ["lt", "gt", "exact"],
    }


class DeviceSetupShapeFilterSet(FilterSet):
    """DeviceSetupShapeFilterSet class."""

    model = DeviceSetupShape
    fields = {
        "name": ["exact", "contains"],
        "description": ["exact", "contains"],
    }


class DeviceSetupFilterSet(FilterSet):
    """DeviceSetupFilterSet class."""

    model = DeviceSetup
    fields = {
        "name": ["exact", "contains"],
        "name_display": ["exact", "contains"],
        "manufacturer__name": ["exact", "contains"],
        "url": ["exact", "contains"],
        "pid": ["exact", "contains"],
        "iri": ["exact", "contains"],
        "model_no": ["exact", "contains"],
        "product_type": ["exact", "contains"],
        "type_barcode": ["exact", "contains"],
        "product_no": ["exact", "contains"],
        "weight": ["lt", "gt", "exact", "range"],
        "energy_consumption": ["lt", "gt", "exact", "range"],
        "datetime_last_modified": ["lt", "gt", "exact", "range"],
        "description": ["exact", "contains"],
    }


class LabwareClassFilterSet(FilterSet):
    """LabwareClassFilterSet class."""

    model = LabwareClass
    fields = {
        "name": ["exact", "contains"],
        "description": ["exact", "contains"],
    }


class WellShapeFilterSet(FilterSet):
    """WellShapeFilterSet class."""

    model = WellShape
    fields = {
        "name": ["exact", "contains"],
        "description": ["exact", "contains"],
    }


class LabwareShapeFilterSet(FilterSet):
    """LabwareShapeFilterSet class."""

    model = LabwareShape
    fields = {
        "name": ["exact", "contains"],
        "well_dist_row": ["lt", "gt", "exact"],
        "well_dist_col": ["lt", "gt", "exact"],
        "description": ["exact", "contains"],
    }


class LabwareFilterSet(FilterSet):
    """LabwareFilterSet class."""

    model = Labware
    fields = {
        "name": ["exact", "contains"],
        "name_display": ["exact", "contains"],
        "manufacturer__name": ["exact", "contains"],
        "url": ["exact", "contains"],
        "pid": ["exact", "contains"],
        "iri": ["exact", "contains"],
        "model_no": ["exact", "contains"],
        "product_type": ["exact", "contains"],
        "type_barcode": ["exact", "contains"],
        "product_no": ["exact", "contains"],
        "weight": ["lt", "gt", "exact", "range"],
        "energy_consumption": ["lt", "gt", "exact", "range"],
        "datetime_last_modified": ["lt", "gt", "exact", "range"],
        "description": ["exact", "contains"],
        "num_rows": ["lt", "gt", "exact"],
        "num_cols": ["lt", "gt", "exact"],
        "num_wells": ["lt", "gt", "exact"],
        #'labware_class': ['exact'],
    }



"""
lara_django_material forms package.

This package contains all forms organized into sub-modules:
- base: Common utilities (CustomFormRenderer, SearchableSelectMultiple),
        ExtraData forms, PartDeviceClass forms
- parts: Part, PartShape forms
- devices: Device, DeviceShape forms
- device_setups: DeviceSetup, DeviceSetupShape forms
- labware: Labware, LabwareClass, LabwareShape, WellShape forms
"""

from .base import (
    CustomFormRenderer,
    ExtraDataCreateForm,
    ExtraDataUpdateForm,
    PartDeviceClassCreateForm,
    PartDeviceClassUpdateForm,
    SearchableSelectMultiple,
)
from .device_setups import (
    DeviceSetupCreateForm,
    DeviceSetupShapeCreateForm,
    DeviceSetupShapeUpdateForm,
    DeviceSetupUpdateForm,
)
from .devices import (
    DeviceCreateForm,
    DeviceShapeCreateForm,
    DeviceShapeUpdateForm,
    DeviceUpdateForm,
)
from .labware import (
    LabwareClassCreateForm,
    LabwareClassUpdateForm,
    LabwareCreateForm,
    LabwareShapeCreateForm,
    LabwareShapeUpdateForm,
    LabwareUpdateForm,
    WellShapeCreateForm,
    WellShapeUpdateForm,
)
from .parts import (
    PartCreateForm,
    PartShapeCreateForm,
    PartShapeUpdateForm,
    PartUpdateForm,
)

__all__ = [
    "CustomFormRenderer",
    "DeviceCreateForm",
    "DeviceSetupCreateForm",
    "DeviceSetupShapeCreateForm",
    "DeviceSetupShapeUpdateForm",
    "DeviceSetupUpdateForm",
    "DeviceShapeCreateForm",
    "DeviceShapeUpdateForm",
    "DeviceUpdateForm",
    "ExtraDataCreateForm",
    "ExtraDataUpdateForm",
    "LabwareClassCreateForm",
    "LabwareClassUpdateForm",
    "LabwareCreateForm",
    "LabwareShapeCreateForm",
    "LabwareShapeUpdateForm",
    "LabwareUpdateForm",
    "PartCreateForm",
    "PartDeviceClassCreateForm",
    "PartDeviceClassUpdateForm",
    "PartShapeCreateForm",
    "PartShapeUpdateForm",
    "PartUpdateForm",
    "SearchableSelectMultiple",
    "WellShapeCreateForm",
    "WellShapeUpdateForm",
]



"""
Parts models for lara_django_material.

This module contains Part-related models including shape, Part,
and subset classes.
"""

import uuid

from django.conf import settings
from django.db import models
from django.utils import timezone
from lara_django_people.models import ItemSubsetAbstr

from .base import ExtraData, PartAbstr, PartDeviceClass, PartShapeAbstr


class PartShape(PartShapeAbstr):
    """Part shape and dimensions."""

    model_curi = "lara:material/PartShape"
    partshape_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    data_extra = models.ManyToManyField(
        ExtraData,
        related_name="%(app_label)s_%(class)s_ps_data_extra_related",
        related_query_name="%(app_label)s_%(class)s_ps_data_extra_related_query",
        blank=True,
        help_text="part shape extra data, like 3D model or polygon describing the part shape/dimensions",
    )

    def save(self, *args, **kwargs):
        """Generate default values for IRI."""
        if self.iri is None or self.iri == "":
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}material/PartShape/{self.name.replace(' ', '_').lower().strip()}"
            )

        super().save(*args, **kwargs)


class Part(PartAbstr):
    """Generic Part class."""

    model_curi = "lara:material/Part"
    part_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    part_class = models.ForeignKey(
        PartDeviceClass,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        help_text="screw, bolt, nut, lamp, diode, resitor, IC, ...",
    )
    shape = models.ForeignKey(
        PartShape,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        help_text="shape of Part",
    )
    components = models.ManyToManyField(
        "self",
        related_query_name="%(app_label)s_%(class)s_components_related_query",
        blank=True,
        help_text="components of the part",
    )
    data_extra = models.ManyToManyField(
        ExtraData,
        related_name="%(app_label)s_%(class)s_parts_data_extra_related",
        related_query_name="%(app_label)s_%(class)s_parts_data_extra_related_query",
        blank=True,
        help_text="e.g. manual",
    )

    def save(self, *args, **kwargs):
        """Generate default values for type barcode and IRI."""
        if self.name is None or self.name == "":
            if self.name_display is not None or self.name_display != "":
                self.name = self.name_display.replace(" ", "_").lower().strip()

        if self.name is not None or self.name != "":
            self.name = self.name.replace(" ", "_").lower().strip()

        if self.name_display is None or self.name_display == "":
            self.name_display = self.name

        if self.type_barcode == "" or self.type_barcode is None:
            self.type_barcode = f"AUTO_{str(uuid.uuid4())[:8]}"  # avoiding unique constraint violation

        if self.iri is None or self.iri == "":
            self.iri = f"{settings.LARA_PREFIXES['lara']}/material/Part/{self.name.replace(' ', '_').lower().strip()}_{str(self.part_id)[:8]}"

        if self.datetime_last_modified is None:
            self.datetime_last_modified = timezone.now()

        super().save(*args, **kwargs)


class PartsSubset(ItemSubsetAbstr):
    """
    Parts subset.

    It can be used for user or group defined subsets of parts.
    """

    model_curi = "lara:material/PartsSubset"
    partssubset_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    parts = models.ManyToManyField(
        Part,
        related_name="%(app_label)s_%(class)s_parts_subset_related",
        related_query_name="%(app_label)s_%(class)s_parts_subset_related_query",
        blank=True,
        help_text="parts in the subset",
        through="OrderedPartsSubset",
    )


class OrderedPartsSubset(models.Model):
    """
    Through model for PartsSubset and Part relationship.

    This class can be used to store the order of parts in a subset.
    """

    model_curi = "lara:material/OrderedPartsSubset"

    orderedpartssubset_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    part = models.ForeignKey(
        Part,
        related_name="%(app_label)s_%(class)s_part_related",
        related_query_name="%(app_label)s_%(class)s_part",
        on_delete=models.CASCADE,
        help_text="part in the subset",
    )

    parts_subset = models.ForeignKey(
        PartsSubset,
        related_name="%(app_label)s_%(class)s_parts_subset_related",
        related_query_name="%(app_label)s_%(class)s_parts_subset",
        on_delete=models.CASCADE,
        help_text="subset of parts",
    )

    order = models.IntegerField(help_text="order of part in subset")

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["part", "parts_subset"],
                name="%(app_label)s_%(class)s_part_parts_subset_unique",
            ),
        ]
        ordering = ("order",)



"""
_____________________________________________________________________.

:PROJECT: LARAsuite

*lara_django_material app *

:details: lara_django_material app configuration.
         This provides a generic django app configuration mechanism.
         For more details see:
         https://docs.djangoproject.com/en/4.0/ref/applications/
         -
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: -
________________________________________________________________________
"""

from django.apps import AppConfig


class LaraDjangoMaterialConfig(AppConfig):
    """LaraDjangoMaterialConfig class."""

    name = "lara_django_material"
    url_path = "material/"
    # enter a verbose name for your app: lara_django_material here - this will be used in the admin interface
    verbose_name = "LARA-django Material"
    # lara_app_icon = 'lara_django_material_icon.svg'  # this will be used to display an icon, e.g. in the main LARA menu.
    lara_app_color = "amber"  # this will be used to highlight the app in the main LARA sidebar and app page
    lara_app_icon = "lara_material_icon.svg"
    verbose_name_short = "Material"
    lara_app_description = "Parts, Devices, Device Setups and Labware"

    def ready(self):
        """Ready."""
        # add urlpatterns
        from django.urls import include, path
        from lara_django.urls import urlpatterns

        urlpatterns += [
            path(self.url_path, include("lara_django_material.urls", namespace="lara_django_material")),
        ]



"""
_____________________________________________________________________.

:PROJECT: LARAsuite

*lara_django_material urls *

:details: lara_django_material urls module.
         - add app specific urls here
         -
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: -
________________________________________________________________________
"""

from django.urls import path
from django_socio_grpc.settings import grpc_settings

from . import views

# Add your {cookiecutter.project_slug}} urls here.


# !! this sets the apps namespace to be used in the template
app_name = "lara_django_material"

# companies and institutions should also be added
# the 'name' attribute is used in templates to address the url independent of the view


urlpatterns = [
    path("extradata/list/", views.ExtraDataSingleTableView.as_view(), name="extradata-list"),
    path("extradata/create/", views.ExtraDataCreateView.as_view(), name="extradata-create"),
    path("extradata/update/<uuid:pk>", views.ExtraDataUpdateView.as_view(), name="extradata-update"),
    path("extradata/delete/<uuid:pk>", views.ExtraDataDeleteView.as_view(), name="extradata-delete"),
    path("extradata/<uuid:pk>/", views.ExtraDataDetailView.as_view(), name="extradata-detail"),
    path("partdeviceclass/list/", views.PartDeviceClassSingleTableView.as_view(), name="partdeviceclass-list"),
    path("partdeviceclass/create/", views.PartDeviceClassCreateView.as_view(), name="partdeviceclass-create"),
    path("partdeviceclass/update/<uuid:pk>", views.PartDeviceClassUpdateView.as_view(), name="partdeviceclass-update"),
    path("partdeviceclass/delete/<uuid:pk>", views.PartDeviceClassDeleteView.as_view(), name="partdeviceclass-delete"),
    path("partdeviceclass/<uuid:pk>/", views.PartDeviceClassDetailView.as_view(), name="partdeviceclass-detail"),
    path("partshape/list/", views.PartShapeSingleTableView.as_view(), name="partshape-list"),
    path("partshape/create/", views.PartShapeCreateView.as_view(), name="partshape-create"),
    path("partshape/update/<uuid:pk>", views.PartShapeUpdateView.as_view(), name="partshape-update"),
    path("partshape/delete/<uuid:pk>", views.PartShapeDeleteView.as_view(), name="partshape-delete"),
    path("partshape/<uuid:pk>/", views.PartShapeDetailView.as_view(), name="partshape-detail"),
    path("part/list/", views.PartSingleTableView.as_view(), name="part-list"),
    path("part/create/", views.PartCreateView.as_view(), name="part-create"),
    path("part/update/<uuid:pk>", views.PartUpdateView.as_view(), name="part-update"),
    path("part/delete/<uuid:pk>", views.PartDeleteView.as_view(), name="part-delete"),
    path("part/<uuid:pk>/", views.PartDetailView.as_view(), name="part-detail"),
    path("device/list/", views.DeviceSingleTableView.as_view(), name="device-list"),
    path("device/create/", views.DeviceCreateView.as_view(), name="device-create"),
    path("device/update/<uuid:pk>", views.DeviceUpdateView.as_view(), name="device-update"),
    path("device/delete/<uuid:pk>", views.DeviceDeleteView.as_view(), name="device-delete"),
    path("device/<uuid:pk>/", views.DeviceDetailView.as_view(), name="device-detail"),
    path("devicesetup/list/", views.DeviceSetupSingleTableView.as_view(), name="devicesetup-list"),
    path("devicesetup/create/", views.DeviceSetupCreateView.as_view(), name="devicesetup-create"),
    path("devicesetup/update/<uuid:pk>", views.DeviceSetupUpdateView.as_view(), name="devicesetup-update"),
    path("devicesetup/delete/<uuid:pk>", views.DeviceSetupDeleteView.as_view(), name="devicesetup-delete"),
    path("devicesetup/<uuid:pk>/", views.DeviceSetupDetailView.as_view(), name="devicesetup-detail"),
    path("labwareclass/list/", views.LabwareClassSingleTableView.as_view(), name="labwareclass-list"),
    path("labwareclass/create/", views.LabwareClassCreateView.as_view(), name="labwareclass-create"),
    path("labwareclass/update/<uuid:pk>", views.LabwareClassUpdateView.as_view(), name="labwareclass-update"),
    path("labwareclass/delete/<uuid:pk>", views.LabwareClassDeleteView.as_view(), name="labwareclass-delete"),
    path("labwareclass/<uuid:pk>/", views.LabwareClassDetailView.as_view(), name="labwareclass-detail"),
    path("wellshape/list/", views.WellShapeSingleTableView.as_view(), name="wellshape-list"),
    path("wellshape/create/", views.WellShapeCreateView.as_view(), name="wellshape-create"),
    path("wellshape/update/<uuid:pk>", views.WellShapeUpdateView.as_view(), name="wellshape-update"),
    path("wellshape/delete/<uuid:pk>", views.WellShapeDeleteView.as_view(), name="wellshape-delete"),
    path("wellshape/<uuid:pk>/", views.WellShapeDetailView.as_view(), name="wellshape-detail"),
    path("labwareshape/list/", views.LabwareShapeSingleTableView.as_view(), name="labwareshape-list"),
    path("labwareshape/create/", views.LabwareShapeCreateView.as_view(), name="labwareshape-create"),
    path("labwareshape/update/<uuid:pk>", views.LabwareShapeUpdateView.as_view(), name="labwareshape-update"),
    path("labwareshape/delete/<uuid:pk>", views.LabwareShapeDeleteView.as_view(), name="labwareshape-delete"),
    path("labwareshape/<uuid:pk>/", views.LabwareShapeDetailView.as_view(), name="labwareshape-detail"),
    path("labware/list/", views.LabwareSingleTableView.as_view(), name="labware-list"),
    path("labware/create/", views.LabwareCreateView.as_view(), name="labware-create"),
    path("labware/update/<uuid:pk>", views.LabwareUpdateView.as_view(), name="labware-update"),
    path("labware/delete/<uuid:pk>", views.LabwareDeleteView.as_view(), name="labware-delete"),
    path("labware/<uuid:pk>/", views.LabwareDetailView.as_view(), name="labware-detail"),
    path("", views.PartSingleTableView.as_view(), name="material-root"),
]

# register handlers in settings
grpc_settings.user_settings["GRPC_HANDLERS"] += ["lara_django_material.grpc.handlers.grpc_handlers"]



"""
_____________________________________________________________________.

:PROJECT: LARAsuite

*lara_django_material Model Context Protocol (MCP) handler *

:details: lara_django_material Model Context Protocol (MCP) handler.

:authors: mark doerr <mark.doerr@uni-greifswald.de>

________________________________________________________________________
"""

from mcp_server import ModelQueryToolset

from .models import Device, ExtraData, Labware, LabwareClass, LabwareShape, LabwareSubset, Part, PartDeviceClass


class ExtradataQueryTool(ModelQueryToolset):
    """ExtradataQueryTool class."""

    model = ExtraData


class PartDeviceClassQueryTool(ModelQueryToolset):
    """PartDeviceClassQueryTool class."""

    model = PartDeviceClass


class DeviceClassQueryTool(ModelQueryToolset):
    """DeviceClassQueryTool class."""

    model = Device


class PartQueryTool(ModelQueryToolset):
    """PartQueryTool class."""

    model = Part


class LabwareClassQueryTool(ModelQueryToolset):
    """LabwareClassQueryTool class."""

    model = LabwareClass


class LabwareQueryTool(ModelQueryToolset):
    """LabwareQueryTool class."""

    model = Labware


class LabwareShapeQueryTool(ModelQueryToolset):
    """LabwareShapeQueryTool class."""

    model = LabwareShape


class LabwareSubsetQueryTool(ModelQueryToolset):
    """LabwareSubsetQueryTool class."""

    model = LabwareSubset


#     def get_queryset(self):
#         """
#         Returns a queryset of all birds with a location.

#         This is used to filter the queryset for the admin interface.
#         self.request can be used to filter the queryset
#         """
#         return super().get_queryset().filter(location__isnull=False)



"""
Labware models for lara_django_material.

This module contains Labware-related models including LabwareClass, WellShape,
LabwareShape, Labware, and subset classes.
"""

import uuid

from django.conf import settings
from django.db import models
from django.utils import timezone
from lara_django_people.models import ItemSubsetAbstr

from .base import ExtraData, PartAbstr, PartShapeAbstr
from .parts import Part


class LabwareClass(models.Model):
    """
    Labware classes.

    Examples: microtiterplate / MTP, tube, flask, ...
    """

    model_curi = "lara:material/LabwareClass"
    labwareclass_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.TextField(unique=True)
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    description = models.TextField(
        default="no description",
        null=True,
        blank=True,
        help_text="description of the labware class",
    )

    class Meta:
        verbose_name_plural = "LabwareClasses"
        # db_table = "lara_labwares_labware_class"

    def __str__(self):
        return (self.name) or ""

    def save(self, *args, **kwargs):
        """Generate default values for IRI."""
        if self.iri is None or self.iri == "":
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}material/LabwareClass/{self.name.replace(' ', '_').lower().strip()}"
            )
        super().save(*args, **kwargs)


class WellShape(models.Model):
    """
    Labware Well Shape.

    Describes shapes of wells: round, square, buffeled, flowered, ...
    This can also be used to describe the bottom shape, like flat, conical.
    """

    model_curi = "lara:material/WellShape"
    wellshape_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.TextField(help_text="Well shape name")
    uri = models.URLField(
        blank=True,
        null=True,
        help_text="Universal Resource Indentifier - URI - can be used to identify the shape ",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    depth_well = models.FloatField(blank=True, null=True, help_text="total well depth=hight")
    shape_well = models.TextField(
        unique=True,
        null=True,
        help_text="labware well shape,e.g. round, square, buffeled,...",
    )
    shape_bottom = models.TextField(unique=True, null=True, help_text="well, bottom shape, flat, round, conical")
    top_radius_xy = models.FloatField(blank=True, null=True, help_text="radius of a round well at the top opening")
    bottom_radius_xy = models.FloatField(blank=True, null=True, help_text="radius of a round bottom in xy direction")
    bottom_radius_z = models.FloatField(
        blank=True,
        null=True,
        help_text="radius of a round bottom in z (hight) direction",
    )
    cone_angle = models.FloatField(blank=True, null=True, help_text="opening angle of cone in deg")
    cone_depth = models.FloatField(blank=True, null=True, help_text="opening angle of cone in deg")
    shape_polygon_xy = models.TextField(
        null=True,
        help_text="generalized shape polygon for more complex well shapes, in xy direction",
    )
    shape_polygon_z = models.TextField(
        null=True,
        help_text="generalized shape polygon for more complex well shapes, in z direction",
    )
    model_2d = models.TextField(blank=True, null=True, help_text="2D model of well shape in, e.g., SVG format")
    model_3d = models.TextField(blank=True, null=True, help_text="3D model of well shape in, e.g., STL format")
    model_json = models.JSONField(blank=True, null=True, help_text="2D or 3D model of Well shape")
    data_extra = models.ManyToManyField(
        ExtraData,
        related_name="%(app_label)s_%(class)s_data_extra_related",
        related_query_name="%(app_label)s_%(class)s_data_extra_related_query",
        blank=True,
        help_text="well shape extra data, like 3D model or polygon describing the well shape/dimensions",
    )
    description = models.TextField(blank=True, null=True, help_text="description of labware shape")

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return self.name or ""

    def save(self, *args, **kwargs):
        """Generate default values for IRI."""
        if self.iri is None or self.iri == "":
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}material/WellShape/{self.name.replace(' ', '_').lower().strip()}"
            )

        super().save(*args, **kwargs)


class LabwareShape(PartShapeAbstr):
    """
    Labware and Lid shape and dimensions.

    E.g., special characteristics of a labware, like material, shape,
    well-shape ...
    """

    model_curi = "lara:material/LabwareShape"
    labwareshape_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    well_dist_row = models.FloatField(blank=True, null=True, help_text="well-to-well distance in row direction")
    well_dist_col = models.FloatField(blank=True, null=True, help_text="well-to-well distance in column direction")
    well_shape = models.ForeignKey(
        WellShape,
        related_name="%(app_label)s_%(class)s_well_shape_related",
        related_query_name="%(app_label)s_%(class)s_well_shape_related_query",
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        help_text="Well shape description",
    )
    data_extra = models.ManyToManyField(
        ExtraData,
        related_name="%(app_label)s_%(class)s_lws_data_extra_related",
        related_query_name="%(app_label)s_%(class)s_lws_data_extra_related_query",
        blank=True,
        help_text="labware shape extra data, like 3D model or polygon describing the labware shape/dimensions",
    )
    description = models.TextField(blank=True, null=True, help_text="description of labware shape")

    def save(self, *args, **kwargs):
        """Generate default values for IRI."""
        if self.iri is None or self.iri == "":
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}material/LabwareShape/{self.name.replace(' ', '_').lower().strip()}"
            )
        super().save(*args, **kwargs)


class Labware(PartAbstr):
    """
    Generic labware class.

    Models/describes the general properties of a labware like geometry, volume,
    manufacturer. All concrete instances shall appear in the labware store.
    """

    model_curi = "lara:material/Labware"
    labware_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    labware_class = models.ForeignKey(
        LabwareClass,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        help_text="class of the labware, e.g. MTP-SBS/SBS, vial/HPLC,  ...",
    )
    num_rows = models.IntegerField(null=True, blank=True, help_text="number of rows of labware")
    num_cols = models.IntegerField(null=True, blank=True, help_text="number of columns of labware")
    num_wells = models.IntegerField(null=True, blank=True, help_text="number of wells - could be auto generated")
    height = models.FloatField(null=True, blank=True, help_text="height of labware in mm")
    shape = models.ForeignKey(
        LabwareShape,
        related_name="%(app_label)s_%(class)s_labware_shape_related",
        related_query_name="%(app_label)s_%(class)s_labware_shape_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="shape of labware",
    )
    components = models.ManyToManyField(
        Part,
        related_query_name="%(app_label)s_%(class)s_components_related_query",
        blank=True,
        help_text="components of the labware",
    )
    # material # polymer, properties, like solvent tolerance, transparency, ....

    # lid info
    lids = models.ManyToManyField("self", blank=True, help_text="reference to possible lids")
    liddable = models.BooleanField(null=True, blank=True, help_text="labware is liddable")

    # seal info
    sealable = models.BooleanField(default=False, null=True, help_text="labware is sealable")
    data_extra = models.ManyToManyField(
        ExtraData,
        related_name="%(app_label)s_%(class)s_labware_data_extra_related",
        related_query_name="%(app_label)s_%(class)s_labware_data_extra_related_query",
        blank=True,
        help_text="e.g. manual",
    )

    def save(self, *args, **kwargs):
        """Generate default values for iri / type_barcode."""
        if self.name is None or self.name == "":
            if self.name_display is not None or self.name_display != "":
                self.name = self.name_display.replace(" ", "_").lower().strip()

        if self.name is not None or self.name != "":
            self.name = self.name.replace(" ", "_").lower().strip()

        if self.name_display is None or self.name_display == "":
            self.name_display = self.name

        if self.type_barcode == "" or self.type_barcode is None:
            self.type_barcode = f"AUTO_{str(uuid.uuid4())[:8]}"  # avoiding unique constraint violation

        if self.iri is None or self.iri == "":
            self.iri = f"{settings.LARA_PREFIXES['lara']}material/Labware/{self.name.replace(' ', '_').lower().strip()}_{str(self.labware_id)[:8]}"

        if self.datetime_last_modified is None:
            self.datetime_last_modified = timezone.now()

        super().save(*args, **kwargs)


class LabwareSubset(ItemSubsetAbstr):
    """
    Labware subset.

    It can be used for user or group defined subsets of labware.
    """

    model_curi = "lara:material/LabwareSubset"
    labwaresubset_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    labware = models.ManyToManyField(
        Labware,
        related_name="%(app_label)s_%(class)s_labware_setup_related",
        related_query_name="%(app_label)s_%(class)s_labware_setup_related_query",
        blank=True,
        help_text="labwares in the subset",
        through="OrderedLabwareSubset",
    )


class OrderedLabwareSubset(models.Model):
    """
    Through model for LabwareSubset and Labware relationship.

    This class can be used to store the order of labwares in a subset.
    """

    model_curi = "lara:material/OrderedLabwareSubset"

    orderedlabwaresubset_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    labware = models.ForeignKey(
        Labware,
        related_name="%(app_label)s_%(class)s_labware_related",
        related_query_name="%(app_label)s_%(class)s_labware",
        on_delete=models.CASCADE,
        help_text="labware in the subset",
    )

    labwares_subset = models.ForeignKey(
        LabwareSubset,
        related_name="%(app_label)s_%(class)s_labwares_subset_related",
        related_query_name="%(app_label)s_%(class)s_labwares_subset",
        on_delete=models.CASCADE,
        help_text="subset of labwares",
    )

    order = models.IntegerField(help_text="order of labware in subset")

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["labware", "labwares_subset"],
                name="%(app_label)s_%(class)s_labware_labwares_subset_unique",
            ),
        ]
        ordering = ("order",)



"""
Base abstract models for lara_django_material.

This module contains abstract base classes used by various material models.
"""

import hashlib
import json
from uuid import uuid4

from django.conf import settings
from django.contrib.postgres.indexes import GinIndex
from django.contrib.postgres.search import SearchVectorField
from django.db import models
from django.utils import timezone
from lara_django_base.models import ExternalResource, ExtraDataAbstr, Tag
from lara_django_people.models import Entity


class PartShapeAbstr(models.Model):
    """
    Abstract Part / Device / Labware shape class.

    Describes the shape and dimensions of a part/device/labware,
    e.g. SBS.96.round.fb
    """

    name = models.TextField(
        unique=True,
        null=True,
        help_text="name of shape dimension set, e.g. SBS.96.round.fb",
    )
    name_standard = models.TextField(
        blank=True, null=True, help_text="underlying labware/shape standard, e.g. SBS"
    )  # might need an extra table
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    width = models.FloatField(blank=True, null=True, help_text="cont. total width, ")
    length = models.FloatField(blank=True, null=True, help_text="cont. total length ")
    height_unlidded = models.FloatField(blank=True, null=True, help_text="cont. height, without lid")
    hight_lidded = models.FloatField(blank=True, null=True, help_text="cont. height incl. lid")
    hight_stacked = models.FloatField(blank=True, null=True, help_text="stacking height, unlidded")
    hight_stacked_lidded = models.FloatField(blank=True, null=True, help_text="stacking height lidded")
    model_2d = models.TextField(blank=True, null=True, help_text="2D model of part shape in, e.g., SVG format")
    model_3d = models.TextField(blank=True, null=True, help_text="3D model of part shape in, e.g., STL format")
    model_json = models.JSONField(blank=True, null=True, help_text="2D or 3D model of part shape in JSON format")
    description = models.TextField(blank=True, null=True, help_text="description of part shape")

    class Meta:
        abstract = True

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return self.name or ""


class PartAbstr(models.Model):
    """
    Abstract base class for Parts, Devices, and Labware.

    A Part is an artifact that is used to build a device or perform an experiment
    (screw, nut, bolt, transistor, IC, lamp, diode, ...).

    NOTE: model part/device state in process online, waiting, busy,
    shutting down, off; default: init
    """

    name_display = models.TextField(
        blank=True,
        null=True,
        help_text="human readable name of the part/device/labware for the UI.",
    )
    name = models.TextField(
        blank=True,
        null=True,
        help_text="short part or device name, should not contain whitespaces",
    )
    url = models.URLField(blank=True, null=True, help_text="Universal Resource Locator - URL")
    pid = models.URLField(blank=True, null=True, unique=True, help_text="handle URI")
    # PAC-ID:  https://github.com/ApiniLabs/PAC-ID
    pac_id = models.URLField(
        blank=True,
        null=True,
        unique=True,
        help_text="Publicly Addressable Content IDentifier is a globally unique identifier that operates independently of a central registry.",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    # in case of mulitiple manufacturer, this could be modeled with data_extra
    manufacturer = models.ForeignKey(
        Entity,
        related_name="%(app_label)s_%(class)s_manufacturer_related",
        related_query_name="%(app_label)s_%(class)s_manufacturer_related_query",
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        help_text="manufacturer of the part/device/labware",
    )
    model_no = models.TextField(blank=True, null=True, help_text="model or part number (of e.g. manufacturer)")
    product_type = models.TextField(blank=True, null=True, help_text="(manufacturer) product type")
    # ',UNIQUE(barcode) ON CONFLICT IGNORE )'),
    type_barcode = models.TextField(
        blank=True,
        null=True,
        unique=True,
        help_text="barcode of the general part type, not of a specific part !!.",
    )
    type_barcode_json = models.JSONField(
        blank=True,
        null=True,
        help_text="Advanced barcodes in JSON for multiple barcode representations, like QR codes etc.",
    )
    product_no = models.TextField(blank=True, null=True, help_text="manufacturer product number")
    weight = models.FloatField(blank=True, null=True, help_text="part net weight in kg")
    energy_consumption = models.FloatField(null=True, blank=True, help_text="energy consumption in W")
    spec_json = models.JSONField(blank=True, null=True, help_text="part specifications in JSON format")
    hardware_config = models.JSONField(
        blank=True,
        null=True,
        help_text="Configuration of the item hardware setup, like connection schemata, wiring, tubing, ...",
    )

    resources_external = models.ManyToManyField(
        ExternalResource,
        related_name="%(app_label)s_%(class)s_external_resources_related",
        related_query_name="%(app_label)s_%(class)s_external_resources_related_query",
        blank=True,
        help_text="external resources, like literature, websites, manuals, ...",
    )
    icon = models.TextField(blank=True, null=True, help_text="XML/SVG icon/logo/drawing of the part")
    image = models.ImageField(
        upload_to="material/",
        blank=True,
        null=True,
        help_text="rel. path/filename to image",
    )
    tags = models.ManyToManyField(
        Tag,
        related_name="%(app_label)s_%(class)s_tags_related",
        related_query_name="%(app_label)s_%(class)s_tags_related_query",
        blank=True,
        help_text="e.g. manual",
    )
    datetime_last_modified = models.DateTimeField(
        blank=True, null=True, help_text="date and time when data was last modified"
    )
    description = models.TextField(blank=True, null=True, help_text="description of the device")
    search_vector = SearchVectorField(null=True)

    class Meta:
        abstract = True
        constraints = [
            models.UniqueConstraint(
                fields=["name", "manufacturer"],
                name="%(app_label)s_%(class)s_name_manufacturer_unique",
            ),
        ]
        indexes = (GinIndex(fields=["search_vector"]),)

    def __str__(self):
        return self.name or ""

    def __repr__(self):
        return self.name or ""


class PartDeviceClass(models.Model):
    """
    Classes for parts and devices.

    Examples: screw, bolt, robot, spectrometer, HPLC, NMR, ...
    """

    model_curi = "lara:material/PartDeviceClass"
    partdeviceclass_id = models.UUIDField(primary_key=True, default=uuid4, editable=False)
    name = models.TextField(
        unique=True,
        help_text="name of the part/device class, like screw, bolt, robot, spectrometer, HPLC, NMR,",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        help_text="International Resource Identifier - IRI: is used for semantic representation in a related ontology",
    )
    description = models.TextField(blank=True, help_text="description")

    def save(self, *args, **kwargs):
        """Generate default values for IRI."""
        if self.iri is None or self.iri == "":
            self.iri = f"{settings.LARA_PREFIXES['lara']}material/PartDeviceClass/{self.name.replace(' ', '_').lower().strip()}"

        super().save(*args, **kwargs)

    def __str__(self):
        return self.name or ""

    class Meta:
        verbose_name_plural = "PartDeviceClasses"
        db_table = "lara_material_part_device_class"


class ExtraData(ExtraDataAbstr):
    """
    Extra information for material models.

    This class can be used to extend material models by extra information,
    e.g., maintenance plan, maintenance history, incidents history,
    image filename, ...
    """

    model_curi = "lara:material/ExtraData"
    image = models.ImageField(
        upload_to="material/images/",
        blank=True,
        null=True,
        help_text="location room map rel. path/filename to image",
    )
    file = models.FileField(upload_to="material/", blank=True, null=True, help_text="rel. path/filename")

    def save(self, *args, **kwargs):
        """Generate default values for version, hash, name, name_full, and IRI."""
        if self.version is None or self.version == "":
            self.version = "0.0.1"

        if self.hash_sha256 is None:
            if self.data_json is not None:
                to_hash = json.dumps(self.data_json) + str(self.extradata_id)
                self.hash_sha256 = hashlib.sha256(to_hash.encode("utf-8")).hexdigest()
            else:
                self.hash_sha256 = hashlib.sha256(str(self.extradata_id).encode("utf-8")).hexdigest()

        if self.name is None or self.name == "":
            if self.name_display is None or self.name_display == "":
                self.name = "_".join(
                    (
                        timezone.now().strftime("%Y%m%d_%H%M%S"),
                        "data",
                        self.hash_sha256[:8],
                    )
                )
            else:
                self.name = self.name_display.replace(" ", "_").lower().strip()

        else:
            self.name = self.name.replace(" ", "_").lower().strip()
        if self.name_full is None or self.name_full == "":
            if self.namespace is not None and self.version is not None:
                self.name_full = f"{self.namespace.uri}/" + "_".join((self.name.replace(" ", "_"), self.version))
            else:
                self.name_full = "_".join((self.name.replace(" ", "_"), self.version))

        if not self.iri:
            netloc = self.namespace.parse_URI().netloc
            path = self.namespace.parse_URI().path

            self.iri = f"{settings.LARA_PREFIXES['lara']}material/ExtraData/{netloc}{path}/{self.name.replace(' ', '_').lower().strip()}"

        if self.datetime_created is None:
            self.datetime_created = timezone.now()

        if self.datetime_last_modified is None:
            self.datetime_last_modified = timezone.now()

        super().save(*args, **kwargs)



"""
_____________________________________________________________________.

:PROJECT: LARAsuite

*lara_django_material admin *

:details: lara_django_material admin module admin backend configuration.
         -
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: - run "lara-django-dev tables_generator lara_django_material >> tables.py" to update this file
________________________________________________________________________
"""

# django Tests s. https://docs.djangoproject.com/en/4.1/topics/testing/overview/ for lara_django_material
# generated with django-extensions tests_generator  lara_django_material > tests.py (LARA-version)

import django_tables2 as tables
from django.utils.html import format_html

from .models import (
    Device,
    DeviceSetup,
    DeviceSetupShape,
    DeviceShape,
    ExtraData,
    Labware,
    LabwareClass,
    LabwareShape,
    Part,
    PartDeviceClass,
    PartShape,
    WellShape,
)


class ImageColumn(tables.Column):
    """ImageColumn class."""

    def render(self, value):
        """Render."""
        return format_html('<img src="/media/{url}" height="48px", width="48px">', url=value)


class EditColumn(tables.Column):
    """EditColumn class."""

    # add edit boostrap icon to column
    def render(self, value):
        """Render."""
        return format_html(
            '<a href="/lara_django_people/entities/update/{url}"><i class="bi bi-pencil-square"></i></a>',
            url=value,
        )


class ExtraDataTable(tables.Table):
    """ExtraDataTable class."""

    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_material:extradata-detail', [tables.A('pk')]))

    class Meta:
        """Meta class."""

        model = ExtraData

        fields = (
            "data_type",
            "namespace",
            "uri",
            "media_type",
            "iri",
            "url",
            "image",
            "file",
        )


class PartDeviceClassTable(tables.Table):
    """PartDeviceClassTable class."""

    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_material:partdeviceclass-detail', [tables.A('pk')]))

    class Meta:
        """Meta class."""

        model = PartDeviceClass

        fields = ("name", "description")


class PartShapeTable(tables.Table):
    """PartShapeTable class."""

    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_material:partshape-detail', [tables.A('pk')]))

    class Meta:
        """Meta class."""

        model = PartShape

        fields = (
            "name",
            "name_standard",
            "width",
            "length",
            "height_unlidded",
            "hight_lidded",
            "hight_stacked",
            "hight_stacked_lidded",
            "model_2d",
            "model_3d",
            "model_json",
        )


class PartTable(tables.Table):
    """PartTable class."""

    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_material:part-detail', [tables.A('pk')]))

    name = tables.Column(linkify=("lara_django_material:part-detail", [tables.A("pk")]))

    edit = tables.Column(
        empty_values=(),
        verbose_name="Edit",
        default="Edit",
        linkify={
            "viewname": "lara_django_material:part-update",
            "args": [(tables.A("pk"))],
        },
        attrs={"a": {"class": "btn btn-warning btn-sm"}},
    )

    delete = tables.Column(
        empty_values=(),
        verbose_name="Delete",
        default="Delete",
        linkify={
            "viewname": "lara_django_material:part-delete",
            "args": [(tables.A("pk"))],
        },
        attrs={"a": {"class": "btn btn-danger btn-sm"}},
    )

    def render_edit(self):
        """Render edit."""
        return format_html('<i class="bi bi-pencil-square"></i>')

    def render_delete(self):
        """Render delete."""
        return format_html('<i class="bi bi-trash"></i>')

    class Meta:
        """Meta class."""

        model = Part

        fields = (
            "image",
            "part_class",
            "name",
            # "manufacturer",
            "model_no",
            "product_type",
            "product_no",
            "edit",
            "delete",
        )


class DeviceShapeTable(tables.Table):
    """DeviceShapeTable class."""

    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_material:deviceshape-detail', [tables.A('pk')]))

    class Meta:
        """Meta class."""

        model = DeviceShape

        fields = (
            "name",
            "name_standard",
            "width",
            "length",
            "height_unlidded",
            "hight_lidded",
            "hight_stacked",
            "hight_stacked_lidded",
        )


class DeviceTable(tables.Table):
    """DeviceTable class."""

    # adding link to column <column-to-be-linked>
    image = ImageColumn("image")
    name = tables.Column(linkify=("lara_django_material:device-detail", [tables.A("pk")]))

    edit = tables.Column(
        empty_values=(),
        verbose_name="Edit",
        default="Edit",
        linkify={
            "viewname": "lara_django_material:device-update",
            "args": [(tables.A("pk"))],
        },
        attrs={"a": {"class": "btn btn-warning btn-sm"}},
        orderable=False,
    )

    order = tables.Column(
        empty_values=(),
        verbose_name="Order",
        default="Order",
        linkify={
            "viewname": "lara_django_material:device-update",
            "args": [(tables.A("pk"))],
        },
        attrs={"a": {"class": "btn bg-blue-300 btn-sm"}},
        orderable=False,
    )

    delete = tables.Column(
        empty_values=(),
        verbose_name="Delete",
        default="Delete",
        linkify={
            "viewname": "lara_django_material:device-delete",
            "args": [(tables.A("pk"))],
        },
        attrs={"a": {"class": "btn btn-danger btn-sm"}},
        orderable=False,
    )

    def render_order(self):
        """Render order."""
        return format_html('<i class="bi bi-pencil-square"></i>')

    def render_edit(self):
        """Render edit."""
        return format_html('<i class="bi bi-pencil-square"></i>')

    def render_delete(self):
        """Render delete."""
        return format_html('<i class="bi bi-trash"></i>')

    class Meta:
        """Meta class."""

        model = Device

        fields = (
            "image",
            "device_class",
            "name",
            # "manufacturer",
            "model_no",
            "product_type",
            "product_no",
            "energy_consumption",
            "edit",
            "order",
            "delete",
        )


class DeviceSetupShapeTable(tables.Table):
    """DeviceSetupShapeTable class."""

    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_material:devicesetupshape-detail', [tables.A('pk')]))

    class Meta:
        """Meta class."""

        model = DeviceSetupShape

        fields = (
            "name",
            "name_standard",
            "width",
            "length",
            "height_unlidded",
            "hight_lidded",
            "hight_stacked",
            "hight_stacked_lidded",
            "model_2d",
            "model_3d",
            "model_json",
        )


class DeviceSetupTable(tables.Table):
    """DeviceSetupTable class."""

    # adding link to column <column-to-be-linked>
    name = tables.Column(linkify=("lara_django_material:devicesetup-detail", [tables.A("pk")]))

    edit = tables.Column(
        empty_values=(),
        verbose_name="Edit",
        default="Edit",
        linkify={
            "viewname": "lara_django_material:devicesetup-update",
            "args": [(tables.A("pk"))],
        },
        attrs={"a": {"class": "btn btn-warning btn-sm"}},
    )

    delete = tables.Column(
        empty_values=(),
        verbose_name="Delete",
        default="Delete",
        linkify={
            "viewname": "lara_django_material:devicesetup-delete",
            "args": [(tables.A("pk"))],
        },
        attrs={"a": {"class": "btn btn-danger btn-sm"}},
    )

    def render_edit(self):
        """Render edit."""
        return format_html('<i class="bi bi-pencil-square"></i>')

    def render_delete(self):
        """Render delete."""
        return format_html('<i class="bi bi-trash"></i>')

    class Meta:
        """Meta class."""

        model = DeviceSetup

        fields = (
            "image",
            "name",
            "setup_class",
            "url",
            # "manufacturer",
            "model_no",
            "product_type",
            "product_no",
            "energy_consumption",
            "edit",
            "delete",
        )


class LabwareClassTable(tables.Table):
    """LabwareClassTable class."""

    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_material:labwareclass-detail', [tables.A('pk')]))

    edit = tables.Column(
        empty_values=(),
        verbose_name="Edit",
        default="Edit",
        linkify={
            "viewname": "lara_django_material:labwareclass-update",
            "args": [(tables.A("pk"))],
        },
        attrs={"a": {"class": "btn btn-warning btn-sm"}},
    )
    delete = tables.Column(
        empty_values=(),
        verbose_name="Delete",
        default="Delete",
        linkify={
            "viewname": "lara_django_material:labwareclass-delete",
            "args": [(tables.A("pk"))],
        },
        attrs={"a": {"class": "btn btn-danger btn-sm"}},
    )

    def render_edit(self):
        """Render edit."""
        return format_html('<i class="bi bi-pencil-square"></i>')

    def render_delete(self):
        """Render delete."""
        return format_html('<i class="bi bi-trash"></i>')

    class Meta:
        """Meta class."""

        model = LabwareClass

        fields = ("name", "iri", "description")


class WellShapeTable(tables.Table):
    """WellShapeTable class."""

    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_material:wellshape-detail', [tables.A('pk')]))

    class Meta:
        """Meta class."""

        model = WellShape

        fields = (
            "name",
            "uri",
            "iri",
            "depth_well",
            "top_radius_xy",
            "bottom_radius_xy",
            "bottom_radius_z",
            "cone_angle",
            "cone_depth",
        )


class LabwareShapeTable(tables.Table):
    """LabwareShapeTable class."""

    # adding link to column <column-to-be-linked>
    # <column-to-be-linked> = tables.Column(linkify=('lara_django_material:labwareshape-detail', [tables.A('pk')]))

    class Meta:
        """Meta class."""

        model = LabwareShape

        fields = (
            "name",
            "name_standard",
            "width",
            "length",
            "height_unlidded",
            "hight_lidded",
            "hight_stacked",
            "hight_stacked_lidded",
            "well_dist_row",
            "well_dist_col",
            "well_shape",
        )


class LabwareTable(tables.Table):
    """LabwareTable class."""

    # adding link to column <column-to-be-linked>
    name = tables.Column(linkify=("lara_django_material:labware-detail", [tables.A("pk")]))

    edit = tables.Column(
        empty_values=(),
        verbose_name="Edit",
        default="Edit",
        linkify={
            "viewname": "lara_django_material:labware-update",
            "args": [(tables.A("pk"))],
        },
        attrs={"a": {"class": "btn btn-warning btn-sm"}},
    )

    delete = tables.Column(
        empty_values=(),
        verbose_name="Delete",
        default="Delete",
        linkify={
            "viewname": "lara_django_material:labware-delete",
            "args": [(tables.A("pk"))],
        },
        attrs={"a": {"class": "btn btn-danger btn-sm"}},
    )

    def render_edit(self):
        """Render edit."""
        return format_html('<i class="bi bi-pencil-square"></i>')

    def render_delete(self):
        """Render delete."""
        return format_html('<i class="bi bi-trash"></i>')

    class Meta:
        """Meta class."""

        model = Labware

        fields = (
            "icon",
            "labware_class",
            "name",
            "product_type",
            "product_no",
            # "manufacturer",
            "model_no",
            "image",
            "num_rows",
            "num_cols",
            "num_of_wells",
            "liddable",
            "sealable",
            "edit",
            "delete",
        )



"""
_____________________________________________________________________.

:PROJECT: LARAsuite

*lara_django_material views *

:details: lara_django_material views module.
         - add app specific urls here
         -
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: -
________________________________________________________________________
"""

from dataclasses import dataclass, field

from django.shortcuts import get_object_or_404
from django.views.generic import CreateView, DeleteView, DetailView, UpdateView
from django_tables2 import SingleTableView

from .forms import (
    DeviceCreateForm,
    DeviceSetupCreateForm,
    DeviceSetupUpdateForm,
    DeviceUpdateForm,
    ExtraDataCreateForm,
    ExtraDataUpdateForm,
    LabwareClassCreateForm,
    LabwareClassUpdateForm,
    LabwareCreateForm,
    LabwareShapeCreateForm,
    LabwareShapeUpdateForm,
    LabwareUpdateForm,
    PartCreateForm,
    PartDeviceClassCreateForm,
    PartDeviceClassUpdateForm,
    PartShapeCreateForm,
    PartShapeUpdateForm,
    PartUpdateForm,
    WellShapeCreateForm,
    WellShapeUpdateForm,
)
from .models import (
    Device,
    DeviceSetup,
    ExtraData,
    Labware,
    LabwareClass,
    LabwareShape,
    Part,
    PartDeviceClass,
    PartShape,
    WellShape,
)
from .tables import (
    DeviceSetupTable,
    DeviceTable,
    ExtraDataTable,
    LabwareClassTable,
    LabwareShapeTable,
    LabwareTable,
    PartDeviceClassTable,
    PartShapeTable,
    PartTable,
    WellShapeTable,
)


@dataclass
class MaterialMenu:
    """MaterialMenu class."""

    menu_items: list[dict] = field(
        default_factory=lambda: [
            {"name": "Part", "path": "lara_django_material:part-list"},
            {"name": "Part-Store", "path": "lara_django_material_store:part-list"},
            {"name": "Devices", "path": "lara_django_material:device-list"},
            {"name": "Devices-Store", "path": "lara_django_material_store:device-list"},
            {"name": "Labware", "path": "lara_django_material:labware-list"},
            {"name": "Labware-Store", "path": "lara_django_material_store:labware-list"},
        ]
    )


class ExtraDataSingleTableView(SingleTableView):
    """ExtraDataSingleTableView class."""

    model = ExtraData
    table_class = ExtraDataTable

    # fields = ('extradata_id', 'data_type', 'data_json',  'file_type', 'iri', 'URL', 'description', 'image', 'file')

    template_name = "lara_django_material/list.html"
    success_url = "/material/extradata/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "ExtraData - List"
        context["create_link"] = "lara_django_material:extradata-create"
        context["menu_items"] = MaterialMenu().menu_items
        return context


class ExtraDataDetailView(DetailView):
    """ExtraDataDetailView class."""

    model = ExtraData

    template_name = "lara_django_material/extradata_detail.html"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "ExtraData - Details"
        context["update_link"] = "lara_django_material:extradata-update"
        context["list_views"] = [
            "ExtraData",
            "PartDeviceClass",
            "PartShape",
            "Part",
            "Device",
            "DeviceSetup",
            "LabwareClass",
            "WellShape",
            "LabwareShape",
            "Labware",
        ]

        return context


class ExtraDataCreateView(CreateView):
    """ExtraDataCreateView class."""

    model = ExtraData

    template_name = "lara_django_material/create_form.html"
    form_class = ExtraDataCreateForm
    success_url = "/material/extradata/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "ExtraData - Create"
        return context


class ExtraDataUpdateView(UpdateView):
    """ExtraDataUpdateView class."""

    model = ExtraData

    template_name = "lara_django_material/update_form.html"
    form_class = ExtraDataUpdateForm
    success_url = "/material/extradata/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "ExtraData - Update"
        context["delete_link"] = "lara_django_material:extradata-delete"
        return context


class ExtraDataDeleteView(DeleteView):
    """ExtraDataDeleteView class."""

    model = ExtraData

    template_name = "lara_django_material/delete_form.html"
    success_url = "/material/extradata/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "ExtraData - Delete"
        context["delete_link"] = "lara_django_material:extradata-delete"
        return context


class PartDeviceClassSingleTableView(SingleTableView):
    """PartDeviceClassSingleTableView class."""

    model = PartDeviceClass
    table_class = PartDeviceClassTable

    fields = ("name", "description")

    template_name = "lara_django_material/list.html"
    success_url = "/material/partdeviceclass/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "PartDeviceClass - List"
        context["create_link"] = "lara_django_material:partdeviceclass-create"
        context["menu_items"] = MaterialMenu().menu_items
        return context


class PartDeviceClassDetailView(DetailView):
    """PartDeviceClassDetailView class."""

    model = PartDeviceClass

    template_name = "lara_django_material/partdeviceclass_detail.html"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "PartDeviceClass - Details"
        context["update_link"] = "lara_django_material:partdeviceclass-update"
        context["menu_items"] = MaterialMenu().menu_items
        return context


class PartDeviceClassCreateView(CreateView):
    """PartDeviceClassCreateView class."""

    model = PartDeviceClass

    template_name = "lara_django_material/create_form.html"
    form_class = PartDeviceClassCreateForm
    success_url = "/material/partdeviceclass/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "PartDeviceClass - Create"
        return context


class PartDeviceClassUpdateView(UpdateView):
    """PartDeviceClassUpdateView class."""

    model = PartDeviceClass

    template_name = "lara_django_material/update_form.html"
    form_class = PartDeviceClassUpdateForm
    success_url = "/material/partdeviceclass/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "PartDeviceClass - Update"
        context["delete_link"] = "lara_django_material:partdeviceclass-delete"
        return context


class PartDeviceClassDeleteView(DeleteView):
    """PartDeviceClassDeleteView class."""

    model = PartDeviceClass

    template_name = "lara_django_material/delete_form.html"
    success_url = "/material/partdeviceclass/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "PartDeviceClass - Delete"
        context["delete_link"] = "lara_django_material:partdeviceclass-delete"
        return context


class PartShapeSingleTableView(SingleTableView):
    """PartShapeSingleTableView class."""

    model = PartShape
    table_class = PartShapeTable

    # fields = ('name', 'name_standard', 'width', 'length', 'height_unlidded', 'hight_lidded', 'hight_stacked', 'hight_stacked_lidded', 'model', 'description', 'partshape_id')

    template_name = "lara_django_material/list.html"
    success_url = "/material/partshape/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "PartShape - List"
        context["create_link"] = "lara_django_material:partshape-create"
        context["menu_items"] = MaterialMenu().menu_items
        return context


class PartShapeDetailView(DetailView):
    """PartShapeDetailView class."""

    model = PartShape

    template_name = "lara_django_material/partshape_detail.html"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "PartShape - Details"
        context["update_link"] = "lara_django_material:partshape-update"
        context["menu_items"] = MaterialMenu().menu_items
        return context


class PartShapeCreateView(CreateView):
    """PartShapeCreateView class."""

    model = PartShape

    template_name = "lara_django_material/create_form.html"
    form_class = PartShapeCreateForm
    success_url = "/material/partshape/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "PartShape - Create"
        return context


class PartShapeUpdateView(UpdateView):
    """PartShapeUpdateView class."""

    model = PartShape

    template_name = "lara_django_material/update_form.html"
    form_class = PartShapeUpdateForm
    success_url = "/material/partshape/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "PartShape - Update"
        context["delete_link"] = "lara_django_material:partshape-delete"
        return context


class PartShapeDeleteView(DeleteView):
    """PartShapeDeleteView class."""

    model = PartShape

    template_name = "lara_django_material/delete_form.html"
    success_url = "/material/partshape/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "PartShape - Delete"
        context["delete_link"] = "lara_django_material:partshape-delete"
        return context


class PartSingleTableView(SingleTableView):
    """PartSingleTableView class."""

    model = Part
    table_class = PartTable

    # fields = ('name', 'name_full', 'URL', 'pid', 'iri', 'manufacturer', 'model_no', 'product_type', 'type_barcode', 'product_no', 'weight', 'spec_json', 'icon', 'image', 'description', 'part_id', 'part_class', 'shape')

    template_name = "lara_django_material/list.html"
    success_url = "/material/part/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "Part - List"
        context["create_link"] = "lara_django_material:part-create"
        context["menu_items"] = MaterialMenu().menu_items
        return context


class PartDetailView(DetailView):
    """PartDetailView class."""

    model = Part

    template_name = "lara_django_material/part_detail.html"

    def get_context_data(self, **kwargs):
        """Get context data."""
        object = get_object_or_404(Part, pk=self.kwargs["pk"])
        product_dict = {
            "product_type": object.product_type,
            "maunfacturers": object.manufacturer,
            "model_no": object.model_no,
        }

        part_specs_dict = {
            "Spec_JSON": "height - 0.3 m",  # object.spec_json,
            "Weight": object.weight,
            "Energy Consumption": object.energy_consumption,
            "Manual": "manual comes here...",  # object.manual,
            "Service manual": "service manual comes here ..",  # object.service_manual,
            "Quickstart": "quickstart comes here",  # object.quickstart,
            "Brochure": "brochure comes here",  # object.brochure,
        }

        part_links_dict = {
            "url": object.url,
        }

        context = super().get_context_data(**kwargs)
        context["section_title"] = "Part - Details"
        context["update_link"] = "lara_django_material:part-update"
        context["menu_items"] = MaterialMenu().menu_items

        context["product_dict"] = product_dict
        context["part_specs_dict"] = part_specs_dict
        context["part_links_dict"] = part_links_dict
        return context


class PartCreateView(CreateView):
    """PartCreateView class."""

    model = Part

    template_name = "lara_django_material/create_form.html"
    form_class = PartCreateForm
    success_url = "/material/part/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "Part - Create"
        return context


class PartUpdateView(UpdateView):
    """PartUpdateView class."""

    model = Part

    template_name = "lara_django_material/update_form.html"
    form_class = PartUpdateForm
    success_url = "/material/part/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "Part - Update"
        context["delete_link"] = "lara_django_material:part-delete"
        return context


class PartDeleteView(DeleteView):
    """PartDeleteView class."""

    model = Part

    template_name = "lara_django_material/delete_form.html"
    success_url = "/material/part/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "Part - Delete"
        context["delete_link"] = "lara_django_material:part-delete"
        return context


class DeviceSingleTableView(SingleTableView):
    """DeviceSingleTableView class."""

    model = Device
    table_class = DeviceTable

    # fields = ('name', 'name_full', 'URL', 'pid', 'iri', 'manufacturer', 'model_no', 'product_type', 'type_barcode', 'product_no', 'weight', 'spec_json', 'icon', 'image', 'description', 'device_id', 'device_class', 'shape')

    template_name = "lara_django_material/list.html"
    success_url = "/material/device/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "Device - List"
        context["create_link"] = "lara_django_material:device-create"
        context["menu_items"] = MaterialMenu().menu_items
        return context


class DeviceDetailView(DetailView):
    """DeviceDetailView class."""

    model = Device

    template_name = "lara_django_material/device_detail.html"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "Device - Details"
        context["update_link"] = "lara_django_material:device-update"
        context["menu_items"] = MaterialMenu().menu_items
        return context


class DeviceCreateView(CreateView):
    """DeviceCreateView class."""

    model = Device

    template_name = "lara_django_material/create_form.html"
    form_class = DeviceCreateForm
    success_url = "/material/device/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "Device - Create"
        return context


class DeviceUpdateView(UpdateView):
    """DeviceUpdateView class."""

    model = Device

    template_name = "lara_django_material/update_form.html"
    form_class = DeviceUpdateForm
    success_url = "/material/device/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "Device - Update"
        context["delete_link"] = "lara_django_material:device-delete"
        return context


class DeviceDeleteView(DeleteView):
    """DeviceDeleteView class."""

    model = Device

    template_name = "lara_django_material/delete_form.html"
    success_url = "/material/device/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "Device - Delete"
        context["delete_link"] = "lara_django_material:device-delete"
        return context


class DeviceSetupSingleTableView(SingleTableView):
    """DeviceSetupSingleTableView class."""

    model = DeviceSetup
    table_class = DeviceSetupTable

    # fields = ('name', 'name_full', 'URL', 'pid', 'iri', 'manufacturer', 'model_no', 'product_type', 'type_barcode', 'product_no', 'weight', 'spec_json', 'icon', 'image', 'description', 'device_id', 'setup_class', 'blueprint_image', 'blueprint_file_type', 'blueprint_file', 'blueprint_pdf', 'scheme_svg', 'assembly')

    template_name = "lara_django_material/list.html"
    success_url = "7material/devicesetup/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "DeviceSetup - List"
        context["create_link"] = "lara_django_material:devicesetup-create"
        context["menu_items"] = MaterialMenu().menu_items
        return context


class DeviceSetupDetailView(DetailView):
    """DeviceSetupDetailView class."""

    model = DeviceSetup

    template_name = "lara_django_material/devicesetup_detail.html"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "DeviceSetup - Details"
        context["update_link"] = "lara_django_material:devicesetup-update"
        context["menu_items"] = MaterialMenu().menu_items
        return context


class DeviceSetupCreateView(CreateView):
    """DeviceSetupCreateView class."""

    model = DeviceSetup

    template_name = "lara_django_material/create_form.html"
    form_class = DeviceSetupCreateForm
    success_url = "/material/devicesetup/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "DeviceSetup - Create"
        return context


class DeviceSetupUpdateView(UpdateView):
    """DeviceSetupUpdateView class."""

    model = DeviceSetup

    template_name = "lara_django_material/update_form.html"
    form_class = DeviceSetupUpdateForm
    success_url = "/material/devicesetup/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "DeviceSetup - Update"
        context["delete_link"] = "lara_django_material:devicesetup-delete"
        return context


class DeviceSetupDeleteView(DeleteView):
    """DeviceSetupDeleteView class."""

    model = DeviceSetup

    template_name = "lara_django_material/delete_form.html"
    success_url = "/material/devicesetup/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "DeviceSetup - Delete"
        context["delete_link"] = "lara_django_material:devicesetup-delete"
        return context


class LabwareClassSingleTableView(SingleTableView):
    """LabwareClassSingleTableView class."""

    model = LabwareClass
    table_class = LabwareClassTable

    # fields = ('labwareclass_id', 'name', 'description')

    template_name = "lara_django_material/list.html"
    success_url = "/material/labwareclass/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "LabwareClass - List"
        context["create_link"] = "lara_django_material:labwareclass-create"
        context["menu_items"] = MaterialMenu().menu_items
        return context


class LabwareClassDetailView(DetailView):
    """LabwareClassDetailView class."""

    model = LabwareClass

    template_name = "lara_django_material/labwareclass_detail.html"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "LabwareClass - Details"
        context["update_link"] = "lara_django_material:labwareclass-update"
        context["menu_items"] = MaterialMenu().menu_items
        return context


class LabwareClassCreateView(CreateView):
    """LabwareClassCreateView class."""

    model = LabwareClass

    template_name = "lara_django_material/create_form.html"
    form_class = LabwareClassCreateForm
    success_url = "/material/labwareclass/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "LabwareClass - Create"
        return context


class LabwareClassUpdateView(UpdateView):
    """LabwareClassUpdateView class."""

    model = LabwareClass

    template_name = "lara_django_material/update_form.html"
    form_class = LabwareClassUpdateForm
    success_url = "/material/labwareclass/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "LabwareClass - Update"
        context["delete_link"] = "lara_django_material:labwareclass-delete"
        return context


class LabwareClassDeleteView(DeleteView):
    """LabwareClassDeleteView class."""

    model = LabwareClass

    template_name = "lara_django_material/delete_form.html"
    success_url = "/material/labwareclass/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "LabwareClass - Delete"
        context["delete_link"] = "lara_django_material:labwareclass-delete"
        return context


class WellShapeSingleTableView(SingleTableView):
    """WellShapeSingleTableView class."""

    model = WellShape
    table_class = WellShapeTable

    # fields = ('wellshape_id', 'depth_well', 'shape_well', 'shape_botton', 'top_radius_xy', 'bottom_radius_xy', 'bottom_radius_z', 'cone_angle', 'cone_depth', 'shape_polygon_xy', 'shape_polygon_z', 'model')

    template_name = "lara_django_material/list.html"
    success_url = "/material/wellshape/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "WellShape - List"
        context["create_link"] = "lara_django_material:wellshape-create"
        context["menu_items"] = MaterialMenu().menu_items
        return context


class WellShapeDetailView(DetailView):
    """WellShapeDetailView class."""

    model = WellShape

    template_name = "lara_django_material/wellshape_detail.html"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "WellShape - Details"
        context["update_link"] = "lara_django_material:wellshape-update"
        context["menu_items"] = MaterialMenu().menu_items
        return context


class WellShapeCreateView(CreateView):
    """WellShapeCreateView class."""

    model = WellShape

    template_name = "lara_django_material/create_form.html"
    form_class = WellShapeCreateForm
    success_url = "/material/wellshape/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "WellShape - Create"
        return context


class WellShapeUpdateView(UpdateView):
    """WellShapeUpdateView class."""

    model = WellShape

    template_name = "lara_django_material/update_form.html"
    form_class = WellShapeUpdateForm
    success_url = "/material/wellshape/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "WellShape - Update"
        context["delete_link"] = "lara_django_material:wellshape-delete"
        return context


class WellShapeDeleteView(DeleteView):
    """WellShapeDeleteView class."""

    model = WellShape

    template_name = "lara_django_material/delete_form.html"
    success_url = "/material/wellshape/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "WellShape - Delete"
        context["delete_link"] = "lara_django_material:wellshape-delete"
        return context


class LabwareShapeSingleTableView(SingleTableView):
    """LabwareShapeSingleTableView class."""

    model = LabwareShape
    table_class = LabwareShapeTable

    # fields = ('name', 'name_standard', 'width', 'length', 'height_unlidded', 'hight_lidded', 'hight_stacked', 'hight_stacked_lidded', 'model', 'labwareshape_id', 'well_dist_row', 'well_dist_col', 'well_shape', 'description')

    template_name = "lara_django_material/list.html"
    success_url = "/material/labwareshape/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "LabwareShape - List"
        context["create_link"] = "lara_django_material:labwareshape-create"
        context["menu_items"] = MaterialMenu().menu_items
        return context


class LabwareShapeDetailView(DetailView):
    """LabwareShapeDetailView class."""

    model = LabwareShape

    template_name = "lara_django_material/labwareshape_detail.html"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "LabwareShape - Details"
        context["update_link"] = "lara_django_material:labwareshape-update"
        context["menu_items"] = MaterialMenu().menu_items
        return context


class LabwareShapeCreateView(CreateView):
    """LabwareShapeCreateView class."""

    model = LabwareShape

    template_name = "lara_django_material/create_form.html"
    form_class = LabwareShapeCreateForm
    success_url = "/material/labwareshape/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "LabwareShape - Create"
        return context


class LabwareShapeUpdateView(UpdateView):
    """LabwareShapeUpdateView class."""

    model = LabwareShape

    template_name = "lara_django_material/update_form.html"
    form_class = LabwareShapeUpdateForm
    success_url = "/material/labwareshape/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "LabwareShape - Update"
        context["delete_link"] = "lara_django_material:labwareshape-delete"
        return context


class LabwareShapeDeleteView(DeleteView):
    """LabwareShapeDeleteView class."""

    model = LabwareShape

    template_name = "lara_django_material/delete_form.html"
    success_url = "/material/labwareshape/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "LabwareShape - Delete"
        context["delete_link"] = "lara_django_material:labwareshape-delete"
        return context


class LabwareSingleTableView(SingleTableView):
    """LabwareSingleTableView class."""

    model = Labware
    table_class = LabwareTable

    # fields = ('name', 'name_full', 'URL', 'pid', 'iri', 'manufacturer', 'model_no', 'product_type', 'type_barcode', 'product_no', 'weight', 'spec_json', 'icon', 'image', 'description', 'labware_id', 'labware_class', 'num_rows', 'num_cols', 'num_of_wells', 'shape', 'lid', 'liddable', 'sealable')

    template_name = "lara_django_material/list.html"
    success_url = "/material/labware/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "Labware - List"
        context["create_link"] = "lara_django_material:labware-create"
        context["menu_items"] = MaterialMenu().menu_items
        return context


class LabwareDetailView(DetailView):
    """LabwareDetailView class."""

    model = Labware

    template_name = "lara_django_material/labware_detail.html"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "Labware - Details"
        context["update_link"] = "lara_django_material:labware-update"
        context["menu_items"] = MaterialMenu().menu_items
        return context


class LabwareCreateView(CreateView):
    """LabwareCreateView class."""

    model = Labware

    template_name = "lara_django_material/create_form.html"
    form_class = LabwareCreateForm
    success_url = "/material/labware/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "Labware - Create"
        return context


class LabwareUpdateView(UpdateView):
    """LabwareUpdateView class."""

    model = Labware

    template_name = "lara_django_material/update_form.html"
    form_class = LabwareUpdateForm
    success_url = "/material/labware/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "Labware - Update"
        context["delete_link"] = "lara_django_material:labware-delete"
        return context


class LabwareDeleteView(DeleteView):
    """LabwareDeleteView class."""

    model = Labware

    template_name = "lara_django_material/delete_form.html"
    success_url = "/material/labware/list"

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["section_title"] = "Labware - Delete"
        context["delete_link"] = "lara_django_material:labware-delete"
        return context



"""
_____________________________________________________________________.

:PROJECT: LARAsuite

*lara_django_myproj gRPC serializer*

:details: lara_django_myproj gRPC serializer.
         -
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: -
________________________________________________________________________
"""

## generated with django-socio-grpc generateprpcinterface lara_django_material  (LARA-version)

import lara_django_material_grpc.v1.lara_django_material_pb2 as lara_django_material_pb2
from django_socio_grpc import proto_serializers
from lara_django_base.models import (
    DataType,
    ExternalResource,
    MediaType,
    Namespace,
    Tag,
)
from lara_django_people.models import (
    Entity,
    Group,
)
from rest_framework.serializers import PrimaryKeyRelatedField, UUIDField

from lara_django_material.models import (
    Device,
    DeviceSetup,
    DeviceSetupShape,
    DeviceSetupsSubset,
    DeviceShape,
    DevicesSubset,
    ExtraData,
    Labware,
    LabwareClass,
    LabwareShape,
    LabwareSubset,
    Part,
    PartDeviceClass,
    PartShape,
    PartsSubset,
    WellShape,
)


class ExtraDataProtoSerializer(proto_serializers.ModelProtoSerializer):
    """ExtraDataProtoSerializer class."""

    data_type = PrimaryKeyRelatedField(
        queryset=DataType.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    namespace = PrimaryKeyRelatedField(queryset=Namespace.objects.all(), pk_field=UUIDField(format="hex_verbose"))
    media_type = PrimaryKeyRelatedField(
        queryset=MediaType.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )

    class Meta:
        """Meta class."""

        model = ExtraData
        proto_class = lara_django_material_pb2.ExtraDataResponse

        proto_class_list = lara_django_material_pb2.ExtraDataListResponse

        fields = "__all__"  # [data_type', namespace', URI', text', XML', JSON', bin', media_type', IRI', URL', description', image', file']


class PartDeviceClassProtoSerializer(proto_serializers.ModelProtoSerializer):
    """PartDeviceClassProtoSerializer class."""

    class Meta:
        """Meta class."""

        model = PartDeviceClass
        proto_class = lara_django_material_pb2.PartDeviceClassResponse

        proto_class_list = lara_django_material_pb2.PartDeviceClassListResponse

        fields = "__all__"  # [name', description']


class PartDeviceClassByNameProtoSerializer(proto_serializers.ModelProtoSerializer):
    """PartDeviceClassByNameProtoSerializer class."""

    class Meta:
        """Meta class."""

        model = PartDeviceClass
        proto_class = lara_django_material_pb2.PartDeviceClassResponse

        proto_class_list = lara_django_material_pb2.PartDeviceClassListResponse

        fields = "__all__"  # [name', description']


class PartShapeProtoSerializer(proto_serializers.ModelProtoSerializer):
    """PartShapeProtoSerializer class."""

    data_extra = PrimaryKeyRelatedField(
        queryset=ExtraData.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )

    class Meta:
        """Meta class."""

        model = PartShape
        proto_class = lara_django_material_pb2.PartShapeResponse

        proto_class_list = lara_django_material_pb2.PartShapeListResponse

        fields = "__all__"  # [name', name_standard', width', length', height_unlidded', hight_lidded', hight_stacked', hight_stacked_lidded', model_2d', model_3d', model_json', description']


class PartProtoSerializer(proto_serializers.ModelProtoSerializer):
    """PartProtoSerializer class."""

    manufacturer = PrimaryKeyRelatedField(
        queryset=Entity.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        # many=True,
        # required=False, allow_null=True
    )
    part_class = PrimaryKeyRelatedField(
        queryset=PartDeviceClass.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        required=False,
        allow_null=True,
    )
    components = PrimaryKeyRelatedField(
        queryset=Part.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    shape = PrimaryKeyRelatedField(
        queryset=PartShape.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    resources_external = PrimaryKeyRelatedField(
        queryset=ExternalResource.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    tags = PrimaryKeyRelatedField(
        queryset=Tag.objects.all(), pk_field=UUIDField(format="hex_verbose"), many=True, required=False, allow_null=True
    )
    data_extra = PrimaryKeyRelatedField(
        queryset=ExtraData.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )

    class Meta:
        """Meta class."""

        model = Part
        proto_class = lara_django_material_pb2.PartResponse

        proto_class_list = lara_django_material_pb2.PartListResponse

        fields = "__all__"  # [name', name_full', URL', handle', IRI', manufacturer', model_no', product_type', type_barcode', product_no', weight', energy_consumption', spec_json', icon', image', description', part_class', shape']


class PartByNameProtoSerializer(proto_serializers.ModelProtoSerializer):
    """PartByNameProtoSerializer class."""

    manufacturer = PrimaryKeyRelatedField(
        queryset=Entity.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        # many=True
    )
    part_class = PrimaryKeyRelatedField(
        queryset=PartDeviceClass.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        required=False,
        allow_null=True,
    )
    components = PrimaryKeyRelatedField(
        queryset=Part.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    shape = PrimaryKeyRelatedField(
        queryset=PartShape.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    resources_external = PrimaryKeyRelatedField(
        queryset=ExternalResource.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    tags = PrimaryKeyRelatedField(
        queryset=Tag.objects.all(), pk_field=UUIDField(format="hex_verbose"), many=True, required=False, allow_null=True
    )
    data_extra = PrimaryKeyRelatedField(
        queryset=ExtraData.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )

    class Meta:
        """Meta class."""

        model = Part
        proto_class = lara_django_material_pb2.PartResponse

        proto_class_list = lara_django_material_pb2.PartListResponse

        fields = "__all__"  # [name', name_full', URL', handle', IRI', manufacturer', model_no', product_type', type_barcode', product_no', weight', energy_consumption', spec_json', icon', image', description', part_class', shape']


class PartsSubsetProtoSerializer(proto_serializers.ModelProtoSerializer):
    """PartsSubsetProtoSerializer class."""

    namespace = PrimaryKeyRelatedField(queryset=Namespace.objects.all(), pk_field=UUIDField(format="hex_verbose"))
    entity = PrimaryKeyRelatedField(queryset=Entity.objects.all(), pk_field=UUIDField(format="hex_verbose"))
    group = PrimaryKeyRelatedField(
        queryset=Group.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    parts = PrimaryKeyRelatedField(queryset=Part.objects.all(), pk_field=UUIDField(format="hex_verbose"), many=True)
    tags = PrimaryKeyRelatedField(
        queryset=Tag.objects.all(), pk_field=UUIDField(format="hex_verbose"), many=True, required=False, allow_null=True
    )

    class Meta:
        """Meta class."""

        model = PartsSubset
        proto_class = lara_django_material_pb2.PartsSubsetResponse

        proto_class_list = lara_django_material_pb2.PartsSubsetListResponse

        fields = "__all__"


class DeviceShapeProtoSerializer(proto_serializers.ModelProtoSerializer):
    """DeviceShapeProtoSerializer class."""

    data_extra = PrimaryKeyRelatedField(
        queryset=ExtraData.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )

    class Meta:
        """Meta class."""

        model = DeviceShape
        proto_class = lara_django_material_pb2.DeviceShapeResponse

        proto_class_list = lara_django_material_pb2.DeviceShapeListResponse

        fields = "__all__"  # [name', name_standard', width', length', height_unlidded', hight_lidded', hight_stacked', hight_stacked_lidded', model_2d', model_3d', model_json', description']


class DeviceProtoSerializer(proto_serializers.ModelProtoSerializer):
    """DeviceProtoSerializer class."""

    manufacturer = PrimaryKeyRelatedField(
        queryset=Entity.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        # many=True
    )
    device_class = PrimaryKeyRelatedField(
        queryset=PartDeviceClass.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        required=False,
        allow_null=True,
    )
    components = PrimaryKeyRelatedField(
        queryset=Part.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    shape = PrimaryKeyRelatedField(
        queryset=DeviceShape.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    resources_external = PrimaryKeyRelatedField(
        queryset=ExternalResource.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    tags = PrimaryKeyRelatedField(
        queryset=Tag.objects.all(), pk_field=UUIDField(format="hex_verbose"), many=True, required=False, allow_null=True
    )
    data_extra = PrimaryKeyRelatedField(
        queryset=ExtraData.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )

    class Meta:
        """Meta class."""

        model = Device
        proto_class = lara_django_material_pb2.DeviceResponse

        proto_class_list = lara_django_material_pb2.DeviceListResponse

        fields = "__all__"  # [name', name_full', URL', handle', IRI', manufacturer', model_no', product_type', type_barcode', product_no', weight', energy_consumption', spec_json', icon', image', description', device_class', shape']


class DeviceByNameProtoSerializer(proto_serializers.ModelProtoSerializer):
    """DeviceByNameProtoSerializer class."""

    manufacturer = PrimaryKeyRelatedField(
        queryset=Entity.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        # many=True
    )
    device_class = PrimaryKeyRelatedField(
        queryset=PartDeviceClass.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        required=False,
        allow_null=True,
    )
    components = PrimaryKeyRelatedField(
        queryset=Part.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    shape = PrimaryKeyRelatedField(
        queryset=DeviceShape.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    resources_external = PrimaryKeyRelatedField(
        queryset=ExternalResource.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    tags = PrimaryKeyRelatedField(
        queryset=Tag.objects.all(), pk_field=UUIDField(format="hex_verbose"), many=True, required=False, allow_null=True
    )
    data_extra = PrimaryKeyRelatedField(
        queryset=ExtraData.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )

    class Meta:
        """Meta class."""

        model = Device
        proto_class = lara_django_material_pb2.DeviceResponse

        proto_class_list = lara_django_material_pb2.DeviceListResponse

        fields = "__all__"  # [name', name_full', URL', handle', IRI', manufacturer', model_no', product_type', type_barcode', product_no', weight', energy_consumption', spec_json', icon', image', description', device_class', shape']


class DevicesSubsetProtoSerializer(proto_serializers.ModelProtoSerializer):
    """DevicesSubsetProtoSerializer class."""

    namespace = PrimaryKeyRelatedField(queryset=Namespace.objects.all(), pk_field=UUIDField(format="hex_verbose"))
    entity = PrimaryKeyRelatedField(queryset=Entity.objects.all(), pk_field=UUIDField(format="hex_verbose"))
    group = PrimaryKeyRelatedField(
        queryset=Group.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    devices = PrimaryKeyRelatedField(queryset=Device.objects.all(), pk_field=UUIDField(format="hex_verbose"), many=True)
    tags = PrimaryKeyRelatedField(
        queryset=Tag.objects.all(), pk_field=UUIDField(format="hex_verbose"), many=True, required=False, allow_null=True
    )

    class Meta:
        """Meta class."""

        model = DevicesSubset
        proto_class = lara_django_material_pb2.DevicesSubsetResponse

        proto_class_list = lara_django_material_pb2.DevicesSubsetListResponse

        fields = "__all__"


class DeviceSetupShapeProtoSerializer(proto_serializers.ModelProtoSerializer):
    """DeviceSetupShapeProtoSerializer class."""

    data_extra = PrimaryKeyRelatedField(
        queryset=ExtraData.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )

    class Meta:
        """Meta class."""

        model = DeviceSetupShape
        proto_class = lara_django_material_pb2.DeviceSetupShapeResponse

        proto_class_list = lara_django_material_pb2.DeviceSetupShapeListResponse

        fields = "__all__"  # [name', name_standard', width', length', height_unlidded', hight_lidded', hight_stacked', hight_stacked_lidded', model_2d', model_3d', model_json', description']


class DeviceSetupProtoSerializer(proto_serializers.ModelProtoSerializer):
    """DeviceSetupProtoSerializer class."""

    manufacturer = PrimaryKeyRelatedField(
        queryset=Entity.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        # many=True,
        required=False,
        allow_null=True,
    )
    setup_class = PrimaryKeyRelatedField(
        queryset=PartDeviceClass.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        required=False,
        allow_null=True,
    )
    components = PrimaryKeyRelatedField(
        queryset=Device.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    shape = PrimaryKeyRelatedField(
        queryset=DeviceSetupShape.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        required=False,
        allow_null=True,
    )
    blueprint_media_type = PrimaryKeyRelatedField(
        queryset=MediaType.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    resources_external = PrimaryKeyRelatedField(
        queryset=ExternalResource.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    tags = PrimaryKeyRelatedField(
        queryset=Tag.objects.all(), pk_field=UUIDField(format="hex_verbose"), many=True, required=False, allow_null=True
    )
    components = PrimaryKeyRelatedField(
        queryset=Device.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    data_extra = PrimaryKeyRelatedField(
        queryset=ExtraData.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )

    class Meta:
        """Meta class."""

        model = DeviceSetup
        proto_class = lara_django_material_pb2.DeviceSetupResponse

        proto_class_list = lara_django_material_pb2.DeviceSetupListResponse

        fields = "__all__"  # [name', name_full', URL', handle', IRI', manufacturer', model_no', product_type', type_barcode', product_no', weight', energy_consumption', spec_json', icon', image', description', setup_class', shape', blueprint_image', blueprint_media_type', blueprint_file', blueprint_pdf', scheme_svg', assembly']


class DeviceSetupByNameProtoSerializer(proto_serializers.ModelProtoSerializer):
    """DeviceSetupByNameProtoSerializer class."""

    manufacturer = PrimaryKeyRelatedField(
        queryset=Entity.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        # many=True,
        required=False,
        allow_null=True,
    )
    setup_class = PrimaryKeyRelatedField(
        queryset=PartDeviceClass.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        required=False,
        allow_null=True,
    )
    shape = PrimaryKeyRelatedField(
        queryset=DeviceSetupShape.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        required=False,
        allow_null=True,
    )
    blueprint_media_type = PrimaryKeyRelatedField(
        queryset=MediaType.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    resources_external = PrimaryKeyRelatedField(
        queryset=ExternalResource.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    tags = PrimaryKeyRelatedField(
        queryset=Tag.objects.all(), pk_field=UUIDField(format="hex_verbose"), many=True, required=False, allow_null=True
    )
    components = PrimaryKeyRelatedField(
        queryset=Device.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    data_extra = PrimaryKeyRelatedField(
        queryset=ExtraData.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )

    class Meta:
        """Meta class."""

        model = DeviceSetup
        proto_class = lara_django_material_pb2.DeviceSetupResponse

        proto_class_list = lara_django_material_pb2.DeviceSetupListResponse

        fields = "__all__"  # [name', name_full', URL', handle', IRI', manufacturer', model_no', product_type', type_barcode', product_no', weight', energy_consumption', spec_json', icon', image', description', setup_class', shape', blueprint_image', blueprint_media_type', blueprint_file', blueprint_pdf', scheme_svg', assembly']


class DeviceSetupsSubsetProtoSerializer(proto_serializers.ModelProtoSerializer):
    """DeviceSetupsSubsetProtoSerializer class."""

    namespace = PrimaryKeyRelatedField(queryset=Namespace.objects.all(), pk_field=UUIDField(format="hex_verbose"))
    entity = PrimaryKeyRelatedField(queryset=Entity.objects.all(), pk_field=UUIDField(format="hex_verbose"))
    group = PrimaryKeyRelatedField(
        queryset=Group.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    devicesetups = PrimaryKeyRelatedField(
        queryset=DeviceSetup.objects.all(), pk_field=UUIDField(format="hex_verbose"), many=True
    )
    tags = PrimaryKeyRelatedField(
        queryset=Tag.objects.all(), pk_field=UUIDField(format="hex_verbose"), many=True, required=False, allow_null=True
    )

    class Meta:
        """Meta class."""

        model = DeviceSetupsSubset
        proto_class = lara_django_material_pb2.DeviceSetupsSubsetResponse

        proto_class_list = lara_django_material_pb2.DeviceSetupsSubsetListResponse

        fields = "__all__"


class LabwareClassProtoSerializer(proto_serializers.ModelProtoSerializer):
    """LabwareClassProtoSerializer class."""

    class Meta:
        """Meta class."""

        model = LabwareClass
        proto_class = lara_django_material_pb2.LabwareClassResponse

        proto_class_list = lara_django_material_pb2.LabwareClassListResponse

        fields = "__all__"  # [name', IRI', description']


class WellShapeProtoSerializer(proto_serializers.ModelProtoSerializer):
    """WellShapeProtoSerializer class."""

    data_extra = PrimaryKeyRelatedField(
        queryset=ExtraData.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )

    # ExtraDataProtoSerializer(many=True)
    class Meta:
        """Meta class."""

        model = WellShape
        proto_class = lara_django_material_pb2.WellShapeResponse

        proto_class_list = lara_django_material_pb2.WellShapeListResponse

        fields = "__all__"  # [name', URI', IRI', depth_well', shape_well', shape_bottom', top_radius_xy', bottom_radius_xy', bottom_radius_z', cone_angle', cone_depth', shape_polygon_xy', shape_polygon_z', model_2d', model_3d', model_json']


class LabwareShapeProtoSerializer(proto_serializers.ModelProtoSerializer):
    """LabwareShapeProtoSerializer class."""

    well_shape = PrimaryKeyRelatedField(
        queryset=WellShape.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    data_extra = PrimaryKeyRelatedField(
        queryset=ExtraData.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )

    # WellShapeProtoSerializer()
    class Meta:
        """Meta class."""

        model = LabwareShape
        proto_class = lara_django_material_pb2.LabwareShapeResponse

        proto_class_list = lara_django_material_pb2.LabwareShapeListResponse

        fields = "__all__"  # [name', name_standard', width', length', height_unlidded', hight_lidded', hight_stacked', hight_stacked_lidded', model_2d', model_3d', model_json', well_dist_row', well_dist_col', well_shape', description']


class LabwareProtoSerializer(proto_serializers.ModelProtoSerializer):
    """LabwareProtoSerializer class."""

    manufacturer = PrimaryKeyRelatedField(
        queryset=Entity.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        # many=True,
        required=False,
        allow_null=True,
    )
    labware_class = PrimaryKeyRelatedField(
        queryset=LabwareClass.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
    )
    shape = PrimaryKeyRelatedField(
        queryset=LabwareShape.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    resources_external = PrimaryKeyRelatedField(
        queryset=ExternalResource.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    tags = PrimaryKeyRelatedField(
        queryset=Tag.objects.all(), pk_field=UUIDField(format="hex_verbose"), many=True, required=False, allow_null=True
    )
    lids = PrimaryKeyRelatedField(
        queryset=Labware.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    components = PrimaryKeyRelatedField(
        queryset=Part.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    data_extra = PrimaryKeyRelatedField(
        queryset=ExtraData.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )

    class Meta:
        """Meta class."""

        model = Labware
        proto_class = lara_django_material_pb2.LabwareResponse

        proto_class_list = lara_django_material_pb2.LabwareListResponse

        fields = "__all__"  # [name', name_full', URL', handle', IRI', manufacturer', model_no', product_type', type_barcode', product_no', weight', energy_consumption', spec_json', icon', image', description', labware_class', num_rows', num_cols', num_of_wells', shape', liddable', sealable']


class LabwareByNameProtoSerializer(proto_serializers.ModelProtoSerializer):
    """LabwareByNameProtoSerializer class."""

    manufacturer = PrimaryKeyRelatedField(
        queryset=Entity.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        # many=True,
    )
    labware_class = PrimaryKeyRelatedField(
        queryset=LabwareClass.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
    )
    shape = PrimaryKeyRelatedField(
        queryset=LabwareShape.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    resources_external = PrimaryKeyRelatedField(
        queryset=ExternalResource.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    tags = PrimaryKeyRelatedField(
        queryset=Tag.objects.all(), pk_field=UUIDField(format="hex_verbose"), many=True, required=False, allow_null=True
    )
    lids = PrimaryKeyRelatedField(
        queryset=Labware.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )
    data_extra = PrimaryKeyRelatedField(
        queryset=ExtraData.objects.all(),
        pk_field=UUIDField(format="hex_verbose"),
        many=True,
        required=False,
        allow_null=True,
    )

    class Meta:
        """Meta class."""

        model = Labware
        proto_class = lara_django_material_pb2.LabwareResponse

        proto_class_list = lara_django_material_pb2.LabwareListResponse

        fields = "__all__"  # [name', name_full', URL', handle', IRI', manufacturer', model_no', product_type', type_barcode', product_no', weight', energy_consumption', spec_json', icon', image', description', labware_class', num_rows', num_cols', num_of_wells', shape', liddable', sealable']


class LabwareSubsetProtoSerializer(proto_serializers.ModelProtoSerializer):
    """LabwareSubsetProtoSerializer class."""

    namespace = PrimaryKeyRelatedField(queryset=Namespace.objects.all(), pk_field=UUIDField(format="hex_verbose"))
    entity = PrimaryKeyRelatedField(queryset=Entity.objects.all(), pk_field=UUIDField(format="hex_verbose"))
    group = PrimaryKeyRelatedField(
        queryset=Group.objects.all(), pk_field=UUIDField(format="hex_verbose"), required=False, allow_null=True
    )
    labware = PrimaryKeyRelatedField(
        queryset=Labware.objects.all(), pk_field=UUIDField(format="hex_verbose"), many=True
    )
    tags = PrimaryKeyRelatedField(
        queryset=Tag.objects.all(), pk_field=UUIDField(format="hex_verbose"), many=True, required=False, allow_null=True
    )

    class Meta:
        """Meta class."""

        model = LabwareSubset
        proto_class = lara_django_material_pb2.LabwareSubsetResponse

        proto_class_list = lara_django_material_pb2.LabwareSubsetListResponse

        fields = "__all__"



"""
_____________________________________________________________________.

:PROJECT: LARAsuite

*lara_django_myproj gRPC handlers*

:details: lara_django_myproj gRPC handlers.
         -
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: -
________________________________________________________________________
"""

# generated with django-socio-grpc generateprpcinterface lara_django_material  (LARA-version)

# import logging
from django_socio_grpc.services.app_handler_registry import AppHandlerRegistry

from lara_django_material.grpc.services import (
    DeviceService,
    DeviceSetupService,
    DeviceSetupShapeService,
    # DeviceSetupByNameService,
    DeviceSetupsSubsetService,
    DeviceShapeService,
    # DeviceByNameService,
    DevicesSubsetService,
    ExtraDataService,
    LabwareClassService,
    LabwareService,
    LabwareShapeService,
    # LabwareByNameService,
    LabwareSubsetService,
    PartDeviceClassByNameService,
    PartDeviceClassService,
    PartService,
    PartShapeService,
    # PartByNameService,
    PartsSubsetService,
    WellShapeService,
)


def grpc_handlers(server):
    """Grpc handlers."""
    app_registry = AppHandlerRegistry("lara_django_material", server)

    app_registry.get_grpc_module = lambda: "lara_django_material_grpc.v1"

    app_registry.register(ExtraDataService)

    app_registry.register(PartDeviceClassService)

    app_registry.register(PartDeviceClassByNameService)

    app_registry.register(PartShapeService)

    app_registry.register(PartService)

    # app_registry.register(PartByNameService)

    app_registry.register(PartsSubsetService)

    app_registry.register(DeviceShapeService)

    app_registry.register(DeviceService)

    # app_registry.register(DeviceByNameService)

    app_registry.register(DevicesSubsetService)

    app_registry.register(DeviceSetupShapeService)

    app_registry.register(DeviceSetupService)

    # app_registry.register(DeviceSetupByNameService)

    app_registry.register(DeviceSetupsSubsetService)

    app_registry.register(LabwareClassService)

    app_registry.register(WellShapeService)

    app_registry.register(LabwareShapeService)

    app_registry.register(LabwareService)

    # app_registry.register(LabwareByNameService)

    app_registry.register(LabwareSubsetService)
